/* IBM C and C++ Compilers, Version 3.6.6.3 --- sqlccipc.C Wed Aug 14 12:22:13 EDT 2002 (C++) */
/********************************************************************************
*
*  OCO SOURCE MATERIALS
*
*  Source File Name = sqlccipc.C      (%W%)
*
*  Component Name   = SQCC
*
*  Descriptive Name = Inter-Process Communication (IPC) Support functions
*
*  Copyright = 5765-205 (c) Copyright IBM Corp 1994, 1994
*              Licensed Materials - Program Property of IBM
*
*  Status =    Development version - 1st draft
*
*  Function =  This file contains all IPC protocol specific functions.
*              These functions are called by the Common Communication
*              Interface (CCI) functions that provide generic communication
*              services to database protocol layer code.
*
*  List of Functions =  sqlccipcinit
*                       sqlccipcdarihandshake
*                       sqlccipconnr
*                       sqlccipcgetbuffer
*                       sqlccipcfreebuffer
*                       sqlccipcdisc
*                       sqlccipcterm
*                       sqlccipcdariterm
*                       sqlccipcsend
*                       sqlccipcrecv
*                       sqlccipcintrr
*                       sqlccipcgetapplid
*                       sqlccipcputapplid
*                       sqlccGetIPCs
*                       sqlccFreeIPCs
*
*  Operating System =   Common
*
*  Feature Activity:
*  feat #     DATE   WHO DESCRIPTION
*  =====================================================================
*  0011325  05/07/94 tjv Newly developed IPC support                @f11325tjv
*  0011954  03/08/94 SAM Added IPC <-> ASCII functions              @f11954SAM
*  0012523  08/23/94 ltp Move sqlofshr to sqle_run_agent            @d12523ltp
*  0012116  08/08/94 pme Changed interface for sqlccgetbuffer and   @f12116pme
*                        sqlccfreebuffer.
*  0012409  24/08/94 SAM DARI V2 changes                            @f12409SAM
*
*  Defect Activity:
*  defect#   DATE    WHO DESCRIPTION
*  =====================================================================
*  011917   08/09/94 tjv Connect should return an error code which  @d11917tjv
*                        will result in an -1032 sqlcode
*  012153   08/31/94 tjv Agents do not come down when apps die      @d12153tjv
*  013112   09/09/94 mah Fix Q op lengths                           @d13112mah
*  13196    26/09/94 tjv Server receive succeeds when client        @d13196tjv
*                        terminates abnornally                      @d13196tjv
*  14036    03/10/94 tjv 30081 occurs when the data sent exceeds    @d14036tjv
*                        the ASLHEAPSZ value                        @d14036tjv
*  13843    03/10/94 mah Add EDU exist list handler to post appl    @d13843mah
*                        if agent dies abnomrally                   @d13843mah
*  14108    10/05/94 tjv Ensure that ASLHEAPSZ returned by the      @d14108tjv
*                        server is that used at the client          @d14108tjv
*  14425    10/12/94 mah If agent is forced, rc = CLOSED_BY_PARTNER @d14425mah
*  14558    10/14/94 SAM Initialize buffer in sqlccIPCtoASCII       @d14558SAM
*  14521    10/17/94 tjv Enable interrupt support                   @d14521tjv
*  14857    10/20/94 tjv Fix timing hole in packet switching code   @d14857tjv
*  14218    10/24/94 tjv Change return codes to CLOSED_BY_PARTNER   @d14218tjv
*                        in order to get the 1224 sqlcode for       @d14218tjv
*                        local agents
*  15017    10/27/94 SAM Don't invoke AgentExit when running DARI   @d15017SAM
*  15397    10/27/94 mah ipcsend() returns an error if agent gone   @d15397mah
*  15691    11/02/94 tjv Use the new interrupt function pointer     @d15691tjv
*  15910    11/05/94 SAM Don't close synch sem during DARI ccinit   @d15910SAM
*  16619    11/17/94 tjv Rework for I2                              @d16619tjv
*  17343    12/01/94 SAM don't sqlodshr in ipcdariterm              @d17343SAM
*  17444   Dec/02/94 aha Made sqloTQMRequest supports instances     @d17444aha
*  17309    12/05/94 tjv Move sqlogins call back into sqlccinit to  @d17309tjv
*                        that the instanceName field is set on the  @d17309tjv
*                        server.  This field is used for DBMON.     @d17309tjv
*  17984    12/14/94 SAM DARI code review cleanup                   @d17984SAM
*  21159    3/3/95   djh Added some wint support
*  23502    3/153/95 mah Was copying too many bytes when copying    @d23502mah
*                        agent_index to TQM interrupt request,      @d23502mah
*                        resulting in size portion of reply         @d23502mah
*                        buffer being corrupted, causing a -2498    @d23502mah
*                        when trying to read the reply.             @d23502mah
*  23098    3/23/95  tjv Ensure passwords are not traced.           @d23098tjv
*  22959    4/06/95  tjv Fix cleanup problems in ipcinit            @d22959tjv
*  18956    5/09/95  glh Update FFDC calls                          @d18956glh
*  34483    9/08/95  jwp Protect *seterr possible storage overlay   @d34483jwp
*  35497  Sep/28/95  aha Used sqlocshr2() to pass instance name     @d35497aha
*  38558   12/04/95  tjv Only cleanup resources in ipcterm if the   @d38558tjv
*                        resources were successfully allocated.     @d38558tjv
*  40097   02/29/96  cia Add async ODBC support                     @d40097cia
*  46356   06/21/96  tjv In sqlccipcdarihandshare if the SSem       @d46356tjv
*                        routines fail dump the zrc value           @d46356tjv
*  46979   07/21/96  scr Log permitted sem failure after abnormal   @d46979scr
*                        disconnect.                                @d46979scr
*  44244  Aug/15/96  jkl PE MLN Merge                               @d44244jkl
*  61766  Feb/24/97  kcs look up node for ipc in conn_att_node,     @d61766kcs
*                        not my_node_num                            @d61766kcs
*  63024  Mar/24/97  kcs use erglob for nodenum, not krcb           @d63024kcs
*  66222  May/01/97  tjv Fix problem in cleanup code when servers   @d66222tjv
*                        crash and remove all IPCs.                 @d66222tjv
*  68862  Jul/07/97  xun Fix problem in cleanup code in sqlccipcinit@d68862xun
*  74674  Dec/09/97  gal Enable VAC++ compilation on NT             @d74674gal
*  80150  Dec/22/97  tjv Use the lConnAttNode field from sqlerglob  @d80150tjv
*                        sqlccipcintr as with the sqlccipcconr      @d80150tjv
*  80241  Jan/12/98  djh handle disappearing local connection       @d80241djh
*  80385  Jan/15/98  xun Don't memset ipcShmHandle to be 0 in       @d80385xun
*                        sqlccFreeIPCs()                            @d80385xun
*  80418  Jan/16/98  xun Fix Dari for NT                            @d80418xun
*  80941  Jan/28/98  tjv For 80418 only use the timeout mechanism   @d80941tjv
*                        in the init function if we are setting     @d80941tjv
*                        up the link to a DARI                      @d80941tjv
*  80788  17/Mar/98  pj  remove SQL_REENTRANT checks around pGp     @d80788pj
*  89989  09/Jul/98  pj  make semaphore errors logged at inf level  @d88989pj
*  98278  14/Feb/99  mfw Support for multiple object types for libdb2
* 106920   26/Mar/99 jh  admin/service message determination.       @d106920jh
* 111814  04/28/99   bpg add sqlccIPCSignalDariEnd                  @d111814bpg
* 104548  06/25/99   gal VAC++ cleanup                              @d104548gal
* 125200  24/Sep/99  jon Allow timeout to be specified for recv     @d125200jon
* 125700  26/Sep/99  jon Allow gateway concentrator to call ipcinit @d125700jon
*                        a second time to reinitialize connection.  @d125700jon
* 126065  29/Sep/99  jon Changes for Gateway Concentrator           @d126065jon
* 127829  15/Oct/99  jon Add additional flag to determine whether to@d127829jon
*                        complete recv if using timeouts            @d127829jon
* 131103  11/Nov/99  xun Add alreadySynced flag in ccipcinit()      @d131103xun
* 138623  27/Jan/00  swm Change dari timeout to 30 seconds on nt    @d138623swm
* 141610  17/Feb/00  xun Fix hang with syncSem problem in ccipcinit @d141610xun
* 143023  22/Feb/00  swm disconnect from shared memory set in dari  @d143023swm
*                        cleanup routine
*
*  APAR Activity:
*   APAR#   DATE     WHO DESCRIPTION
*  =====================================================================
*   45810  07/Jun/96 tjv Ensure that the send and recv semaphore    @d45810tjv
*                        are not reopened in the agent when reusing @d45810tjv
*                        a DARI process                             @d45810tjv
*   51814  23/Oct/96 tjv Check for FORCE conditions during server   @d51814tjv
*                        side initialization.                       @d51814tjv
*   61259  04/Mar/97 tjv There is no need to perform the UNDO in    @d61259tjv
*                        sqlccipcdarihandshake as it is done in     @d61259tjv
*                        sqlccipcinit.  This leads to us eventually @d61259tjv
*                        exhausting the system UNDO limit.          @d61259tjv
*   119602 Jun-14-99 pj  64BIT: purge use of long                   @d119602pj
*   120677 Jul-07-99 pj  Add casts for sqloSSemP timeout args.
*   154946 Jul-13-00 ltp remove code to set clientPid in sqlccipcconnr
*   190698 Feb-01-02 swm ipc conn id no unique in udb EEE            @d190698swm
*   219059 Mar-28-02 swm spurious probes for keepdari=no             @d219059swm
*
*  Last Changed =    %E%  %U%
*
***********************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "sqlz.h"
#include "sqltcc.h"
#include "sqltcc2.h"                                              //@d111814bpg
#include "sqlo.h"
#include "sqlomem.h"
#include "sqlocall.h"
#include "sqlekrcb.h"
#include "sqlei.h"

#include "sqlcccmn.h"
#include "sqlccipc.h"
#include "sqledari.h"
#include "sqlcreen.h"                                             /*@d13843mah*/
#include "sqlereen.h"                                             //@d51814tjv
                                                                  /*@bd13843mah*/

#include "sqlzpdb.h"                                                /*@d44244jkl*/
#include "sqloreen.h"
#include "sqleprom.h"

#include "sqlelagt.h"                                             //@d125700jon

const char SQLCC_FFDC_UID[]="User ID";                            /*@d18956glh*/
const char SQLCC_FFDC_IPC_SHM_HANDLE[]="IPC Shared Memory Handle";/*@d18956glh*/
const char SQLCC_FFDC_IPC_SHM_SIZE[]="IPC Shared Memory Size";    /*@d18956glh*/
const char SQLCC_FFDC_IPC_SEM_SETID[]="IPC Semaphore Set ID";     /*@d18956glh*/
const char SQLCC_FFDC_RECV_SEMID[]="Receive Semaphore ID";        /*@d18956glh*/
const char SQLCC_FFDC_SEND_SEMID[]="Send Semaphore ID";           /*@d18956glh*/
const char SQLCC_FFDC_SYNC_SEMID[]="Synchronization Semaphore ID";/*@d18956glh*/

/****************************************************************************/
/*  macro for setting protocol specific condition sub-structure             */
/****************************************************************************/
#ifndef BLD_CLIENT_ONLY

#define ipcseterr(retcd,sev,func,sqlo_rc)                                   \
               rc = retcd;                                                  \
               pCondition->sSeverity = sev;                                 \
               pCondition->usNumOfTokens = 7;                               \
               sprintf (pCondition->msgTokens[0],                           \
                        "%d", sqlo_rc );                                    \
               strcpy  (pCondition->msgTokens[1],                           \
                        "*");                                               \
               strcpy  (pCondition->msgTokens[2],                           \
                        "*");                                               \
               strcpy  (pCondition->msgTokens[3],                           \
                        "*");                                               \
               strcpy  (pCondition->msgTokens[4],                           \
                        "*");                                               \
               strcpy  (pCondition->msgTokens[5],                           \
                   ((SQLCC_IPC_CON_HANDLE_T *)                              \
                    pComHandle->connHandle.pConnHandleProto)->commHostId);  \
               strncpy (pCondition->msgTokens[6],            /*@f34483jwp*/ \
                        func,SQLCC_TOKEN_LENGTH)

/*****************************************************************************
*
*  Function Name     = sqlccipcinit
*
*  Descriptive Name  = IPC protocol support initialization
*
*  Func. Prototype   = short sqlccipcinit(
*                             SQLCC_INITSTRUCT_T    *pInitStruct,
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             SQLCC_COND_T          *pCondition
*                                      )
*
*  Function          = This function performs preparation work before
*                      a IPC connection can be set up.  It acquires
*                      all the required IPC specific data structures,
*                      and set up all the linkage between the comm.
*                      handle and those structures, as well as IPC
*                      specific functions to be used later.
*
*                      If the connection handle is already available
*                      (i.e. if called on the server side), this
*                      function also saves it in the comm. handle.
*
*  Input             = 1) initStruct: a structure containing a pointer
*                                     to the Connection Handle which,
*                                     if present, must be saved in the
*                                     Communication Handle.
*
*                      2) pComHandle: a pointer to the Communication
*                                     Handle that contains pointers
*                                     to all the IPC specific
*                                     functions and data structures.
*                                     These pointers were null
*                                     and are set by this function.
*
*  Output            = 1) pCondition: a pointer to the common Condition
*                                     structure that contains a pointer
*                                     to the IPC specific condition
*                                     structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_MEM_UNAVAIL
*                      SQLCC_ERR_COMM_REQUEST_FAILED
*                      SQLCC_ERR_INSUFFICIENT_TARGET_INFO
*
*****************************************************************************/

short sqlccipcinit(
                   SQLCC_INITSTRUCT_T            *pInitStruct,
                   SQLCC_COMHANDLE_T             *pComHandle,
                   SQLCC_COND_T                  *pCondition
                  )
{
   SQLCC_IPC_RESOURCES_T  *ipcResources;
   SQLCC_IPC_ASL_HDR_T    *aslCommHeader = NULL;                    //@d131103xun
   struct sql_static_data *pGp = sqlo_get_static_data();            //@d80418xun
   int                     syncDone = FALSE;                        //@d131103xun
   int                     sqlo_rc = SQLO_OK;
   short                   rc = SQLCC_OK;


   struct  /* Flags to assist in error handling */
   {
      unsigned  cleanup  : 1;
      unsigned  send_sem : 1;
      unsigned  recv_sem : 1;
      unsigned  sync_sem : 1;
      unsigned  con_hand : 1;
      unsigned  asl_heap : 1;                                       /*@f12409SAM*/
      unsigned  freeSync : 1;                                       /*@d22959tjv*/
   } errors = {0};

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCINIT);

   /* Set up function pointers in Comhandle */

   pComHandle->termf         = sqlccipcterm;
   pComHandle->connf         = sqlccipcconnr;
   pComHandle->disconnf      = sqlccipcdisc;
   pComHandle->sendf         = sqlccipcsend;
   pComHandle->recvf         = sqlccipcrecv;
   pComHandle->resetf        = sqlccipcterm;
   pComHandle->intrptf_db2v2 = sqlccipcintrr;                     /*@d15691tjv*/
   pComHandle->getapplidf    = sqlccipcgetapplid;
   pComHandle->putapplidf    = sqlccipcputapplid;
   pComHandle->getbufferf    = sqlccipcgetbuffer;                 /*@f12116pme*/
   pComHandle->freebufferf   = sqlccipcfreebuffer;                /*@f12116pme*/

   /* dariFlag - initialized by a DARI               */
   /* dariReused - DARI reused from the pool         */
   /* dariRunning - Agent initializing to use a DARI */

   if (pInitStruct->dariFlag)                                       /*@f12409SAM*/
   {                                                                /*@f12409SAM*/
     pComHandle->termf    = sqlccipcdariterm;                       /*@f12409SAM*/
     pComHandle->resetf   = sqlccipcdariterm;                       /*@f12409SAM*/
   }                                                                /*@f12409SAM*/

   ipcResources = (SQLCC_IPC_RESOURCES_T *)&pInitStruct->ipcResources;

   /* Allocate IPC specific CONNECTION HANDLE, initialize it */
   /* and then set its pointer in the ComHandle              */

   sqlo_rc = sqlogblk(NULL,
                      sizeof(SQLCC_IPC_CON_HANDLE_T),
                      &pComHandle->connHandle.pConnHandleProto);

   if (sqlo_rc != SQLO_OK )
   {
      sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 1,
                       sizeof(sqlo_rc), (void *) &sqlo_rc);
      rc = SQLCC_ERR_MEM_UNAVAIL;
      goto exit;
   }
   errors.con_hand = TRUE;

   sqlomset(pComHandle->connHandle.pConnHandleProto, 0,
            sizeof(SQLCC_IPC_CON_HANDLE_T));

   /* If this function was invoked from within a DARI, or if invoked   */
   /* by the top level server driver for a DARI conversation, then     */
   /* we are guaranteed that the associated IPC resources have been    */
   /* allocated.  These resources were returned by the sqlccGetIpcInfo */
   /* routine, which also anchored the control block in the SQCC       */
   /* initialization control block.  When we encounter this control    */
   /* block we will anchor the information in the connection handle    */
   /* and attach to the specified resources.  If invoked from a server */
   /* which is initializing for the primary connection, simply setup   */
   /* the connection handle.                                           */

   if ((pInitStruct->dariFlag) ||
       (pInitStruct->server_flag))
   {
       errors.freeSync = TRUE;                                    /*@d22959tjv*/

       if (pInitStruct->server_flag &&                              //@db137385xun
           !pInitStruct->dariRunning)
       {
          // if this is server and not running dari, the agent
          // has already connected to the shared memory set,
          // and thus we can get the aslCommHeader
          aslCommHeader = (SQLCC_IPC_ASL_HDR_T *)
                     sqloGetSetRawAddress(&ipcResources->ipcShmHandle);
          if (aslCommHeader->syncDone)
             syncDone = TRUE;
       }                                                            //@de137385xun

       /* Open the send and recv semaphore if not yet open */     /*@bd45810tjv*/
#ifdef SQLO_HAVE_THREADED_ENGINE                                    //@db137385xun
       // in NT and OS/2, we only need to open the semaphores once
       // then they can be used by all threads
       // syncDone will be turned on the first time calling ccipcinit()
       if (!ipcResources->sendRecvSemOpen &&
           !syncDone)
#else
       /* Open the send and recv semaphore if not yet open */     /*@bd45810tjv*/
       if (!ipcResources->sendRecvSemOpen)
#endif                                                              //@de137385xun
       {
          /* Open the send semaphore */

          sqlo_rc = sqloSSemOpen(&ipcResources->ipcSemSetID,
                                 &ipcResources->sendSemID);
          if (sqlo_rc != SQLO_OK)
          {
             rc = SQLCC_ERR_COMM_REQUEST_FAILED;
             sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 2, 8,
                            (void *)&ipcResources->ipcSemSetID, 8,
                            (void *)&ipcResources->sendSemID);
             ipcseterr(rc,8,"OPEN SEMAPHORE",sqlo_rc);
             errors.cleanup = TRUE;
             goto exit;
          }
          errors.send_sem = TRUE;

          /* Open the receive semaphore */

          sqlo_rc = sqloSSemOpen(&ipcResources->ipcSemSetID,
                                 &ipcResources->recvSemID);
          if (sqlo_rc != SQLO_OK)
          {
             rc = SQLCC_ERR_COMM_REQUEST_FAILED;
             sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 3, 8,
                            (void *)&ipcResources->ipcSemSetID, 8,
                            (void *)&ipcResources->recvSemID);
             ipcseterr(rc,8,"OPEN SEMAPHORE",sqlo_rc);
             errors.cleanup = TRUE;
             goto exit;
          }
          errors.recv_sem = TRUE;
          ipcResources->sendRecvSemOpen = TRUE;
       }                                                         /*@ed45810tjv*/
#ifdef SQLO_HAVE_THREADED_ENGINE                                    //@db137385xun
       else
       {
          errors.send_sem = TRUE;
          errors.recv_sem = TRUE;
          ipcResources->sendRecvSemOpen = TRUE;
       }
#endif                                                              //@de137385xun

       if (syncDone)                                                //@de131103xun
       {
          // syncDone flag may only be set to TRUE on client-server
          // connection. It will NOT be set to TRUE if the ipc is used for DARI
          // communication. So within this if statement, we don't need to
          // check dariRunning flag, we know it is FALSE.
          // when this flag is TRUE, the agent has already syned with the
          // client, so no need to (and should not) perform the sync again.
          pComHandle->connHandle.usConnAllocated =
                         pInitStruct->csInit.sInit.usConnAllocated;

          /* Set up a pointer to the ipc specific connection     *//*@bd13843mah*/
          /* handle in the EDU static data area for this agent.  */
          /* This is needed by the EDU exit list routine which is*/
          /* installed below.                                    */


          // This is not DARI case
          pGp->pstSqlc->pConnHandleProto = pComHandle->connHandle.pConnHandleProto;

          /* This installs an exit list routine which will be invoked     */
          /* when this agent dies abnormally.  (The exit list routine is  */
          /* removed in ipcdisc(), so it doesn't get invoked when the     */
          /* agent dies normally).  The exit list routine tells the client*/
          /* of the agent's abnormal death by turning on the recvTermFlag */
          /* and posting the server's send semaphore.                     */

          sqlo_rc = sqloAddEDUExitList (
#if defined(SQLWINT_VA)                           /*@d104548gal*/   /*@d74674gal*/
                                         (SQLO_EDU_EXIT*)        /*@d74674gal*/
#endif                                                              /*@d74674gal*/
                                         sqlccipcAgentExitList );
          if (sqlo_rc != SQLO_OK)
          {
             sqlt_system_error (SQLT_SQLCC, SQLT_SQLCCIPCINIT,
                                20, sizeof(sqlo_rc), &sqlo_rc);
             errors.cleanup = TRUE;
             rc = SQLCC_ERR_COMM_REQUEST_FAILED;
             goto exit;                                        /*@d22959tjv*/
          }                                                     /*@ed13843mah*/
       }
       else
       {                                                               //@de125700jon
          /* Open the synch semaphore */

          sqlo_rc = sqloSSemOpen(&ipcResources->ipcSemSetID,
                                 &ipcResources->syncSemID);
          if (sqlo_rc != SQLO_OK)
          {
             rc = SQLCC_ERR_COMM_REQUEST_FAILED;
             sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 4, 8,
                            (void *)&ipcResources->ipcSemSetID, 8,
                            (void *)&ipcResources->syncSemID);
             ipcseterr(rc,8,"OPEN SEMAPHORE",sqlo_rc);               /*@d14218tjv*/
             errors.cleanup = TRUE;
             goto exit;
          }
          errors.sync_sem = TRUE;

          pComHandle->connHandle.usConnAllocated =
                         pInitStruct->csInit.sInit.usConnAllocated;

          /* synchronize the semaphores with the client or DARI */

          if (pInitStruct->server_flag)
          {
             /* Post the server's receive semaphore, this will wake up     */
             /* the client which will be waiting on this semaphore.        */

             sqlo_rc = sqloSSemV(&ipcResources->ipcSemSetID,
                                 &ipcResources->recvSemID,
                                 0);

             if (sqlo_rc != SQLO_OK)
             {
                sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 6, 8,
                               (void *)&ipcResources->ipcSemSetID, 8,
                               (void *)&ipcResources->recvSemID);
                rc = SQLCC_ERR_COMM_REQUEST_FAILED;
                errors.cleanup = TRUE;                               /*@d16619tjv*/
                ipcseterr(rc,8,"POST SEMAPHORE",sqlo_rc);            /*@d14218tjv*/
                goto exit;
             }

             /* Wait on the synch semaphore to complete initialization */
             {                                                         //@bd51814tjv
                unsigned char * forced_flag =
                            &pGp->pstSqle->s_sqle_agentcb->forced_flag;

                for (sqlo_rc = SQLO_USER_TERM ;
                     (((sqlo_rc == SQLO_USER_TERM) ||                  //@d80418xun
                       (sqlo_rc == SQLO_SSEM_TIMEOUT)) &&              //@d80418xun
                      !(*forced_flag)) ; )
                {
#if defined(NOKILLHANDLER)                                          //@db80418xun
                   // Only perform the timeout check when initializing   @bd80941tjv
                   // communication with a DARI.  There is no way that
                   // we can ensure the client has setup the PID field
                   // when going onto this call.

                   if (pInitStruct->dariRunning)                       //@ed80941tjv
                   {
                      int tmp_rc;
                      // We are on the server, on this system there is
                      // no way to handle the case where an process has
                      // been killed. Wake up every 30 seconds and determine
                      // if the dari process still exists
                      sqlo_rc = sqloSSemP(&ipcResources->ipcSemSetID,
                                          &ipcResources->syncSemID,
                                          SQLO_SSEM_INTERRUPTIBLE,
                                          30);                     //@d138623swm
                      if (sqlo_rc == SQLO_SSEM_TIMEOUT)
                      {
                         tmp_rc = SQLOQPRC(pGp->pstSqle->s_sqle_agentcb->DariMine->dari_pid);
                         if (tmp_rc == SQLO_PROC_NOT_FOUND)
                         {
                            pInitStruct->dariInitError = TRUE;     //@d138623swm
                            break;
                         }
                      }
                   }                                                   //@bd80941tjv
                   else
                   {
                      // We are on the server, on this system there is
                      // no way to handle the case where an process has
                      // been killed. Wake up every 120 seconds and determine
                      // if the client process still exists
                      sqlo_rc = sqloSSemP(&ipcResources->ipcSemSetID,
                                          &ipcResources->syncSemID,
                                          SQLO_SSEM_INTERRUPTIBLE,
                                          120);                     //@db141610xun
                      if (sqlo_rc == SQLO_SSEM_TIMEOUT)
                      {
                         if (aslCommHeader->clientPid != SQLO_INVALID_PID)
                         {
                            sqlo_rc = SQLOQPRC(aslCommHeader->clientPid);
                            if (sqlo_rc == SQLO_PROC_NOT_FOUND)
                            {
                               break;
                            }
                         }
                      }                                             //@de141610xun
                   }                                                   //@ed80941tjv
#else                                                               //@de80418xun
                   sqlo_rc = sqloSSemP(&ipcResources->ipcSemSetID,
                                       &ipcResources->syncSemID,
                                       SQLO_SSEM_INTERRUPTIBLE,
                                       (U32BIT)SQLO_NO_TIMEOUT);     /*@d120677pj*/
#endif                                                              //@d80418xun
                }
             }                                                         //@ed51814tjv

             if (sqlo_rc != SQLO_OK)
             {
                sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 8, 8,
                               (void *)&ipcResources->ipcSemSetID, 8,
                               (void *)&ipcResources->syncSemID);
                rc = SQLCC_ERR_COMM_REQUEST_FAILED;
                errors.cleanup = TRUE;                               /*@d16619tjv*/
                ipcseterr(rc,8,"SEMAPHORE WAIT",sqlo_rc);            /*@d14218tjv*/
                goto exit;
             }

             /* Close the sync semaphore */

             sqlo_rc = sqloSSemClose(&ipcResources->ipcSemSetID,
                                     &ipcResources->syncSemID);
             errors.sync_sem = FALSE;
             if (sqlo_rc != SQLO_OK)
             {
                sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 9, 8,
                               (void *)&ipcResources->ipcSemSetID, 8,
                               (void *)&ipcResources->syncSemID);
                errors.cleanup = TRUE;
                rc = SQLCC_ERR_COMM_REQUEST_FAILED;
                ipcseterr(rc,8,"SEMAPHORE CLOSE",sqlo_rc);           /*@d14218tjv*/
             }

             /* Free the sync semaphore */

             if (pInitStruct->dariRunning == FALSE)                  /*@d15910SAM*/
             {                                                       /*@d15910SAM*/

               // set the syncDone flag here, if this ipc
               // is used for client-server communication, not
               // dari
               aslCommHeader->syncDone = TRUE;                       //@d131103xun

               errors.freeSync = FALSE;                              /*@d22959tjv*/
               sqlo_rc = sqloSSemSetFree(&ipcResources->ipcSemSetID,
                                         &ipcResources->syncSemID);
               if (sqlo_rc != SQLO_OK)
               {
                  sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 10, 8,
                                 (void *)&ipcResources->ipcSemSetID, 8,
                                 (void *)&ipcResources->syncSemID);
                  errors.cleanup = TRUE;
                  ipcseterr(rc,8,"SEMAPHORE FREE",sqlo_rc);           /*@d14218tjv*/
                  rc = SQLCC_ERR_COMM_REQUEST_FAILED;
               }
             }                                                       /*@d15910SAM*/

             /* If either the close or the free of the synch */     /*@db22959tjv*/
             /* fails go to the cleanup exit point.          */
             if (rc != SQLCC_OK)
             {
                goto exit;
             }                                                      /*@de22959tjv*/

             /* Set up a pointer to the ipc specific connection     *//*@bd13843mah*/
             /* handle in the EDU static data area for this agent.  */
             /* This is needed by the EDU exit list routine which is*/
             /* installed below.                                    */

             /* don't want to overwrite this during the recursive call to    */
             /* sqle_db2ra_as_driver() during the running of a DARI          */

             if (pInitStruct->dariRunning == FALSE)                    /*@d15017SAM*/
             {                                                         /*@d15017SAM*/
               pGp->pstSqlc->pConnHandleProto = pComHandle->connHandle.pConnHandleProto;
             }                                                         /*@d15017SAM*/

             /* This installs an exit list routine which will be invoked     */
             /* when this agent dies abnormally.  (The exit list routine is  */
             /* removed in ipcdisc(), so it doesn't get invoked when the     */
             /* agent dies normally).  The exit list routine tells the client*/
             /* of the agent's abnormal death by turning on the recvTermFlag */
             /* and posting the server's send semaphore.                     */

             if (pInitStruct->dariRunning == FALSE)                    /*@d15017SAM*/
             {                                                         /*@d15017SAM*/
                sqlo_rc = sqloAddEDUExitList (
#if defined(SQLWINT_VA)                           /*@d104548gal*/   /*@d74674gal*/
                                               (SQLO_EDU_EXIT*)        /*@d74674gal*/
#endif                                                              /*@d74674gal*/
                                               sqlccipcAgentExitList );
                if (sqlo_rc != SQLO_OK)
                {
                   sqlt_system_error (SQLT_SQLCC, SQLT_SQLCCIPCINIT,
                                                  11, sizeof(sqlo_rc), &sqlo_rc);
                   errors.cleanup = TRUE;
                   rc = SQLCC_ERR_COMM_REQUEST_FAILED;
                   goto exit;                                        /*@d22959tjv*/
                }                                                     /*@ed13843mah*/
             }                                                         /*@d15017SAM*/
          }
          else     /* executing in the DARI */
          {
             /* Reverse the send and receive semaphores */
             SQLO_SSEMID temp;
             sqlomcpy(&temp, &ipcResources->sendSemID,
                      sizeof(ipcResources->sendSemID));
             sqlomcpy(&ipcResources->sendSemID,
                      &ipcResources->recvSemID,
                      sizeof(ipcResources->recvSemID));
             sqlomcpy(&ipcResources->recvSemID, &temp,
                      sizeof(temp));

             /* Wait on the server's receive semaphore, this will be posted  */
             /* by the server.  This operation is performed to ensure that   */
             /* the dari process performs an UNDO operation on the server's  */
             /* receive semaphore.  This will ensure that the server is      */
             /* posted if the dari process exits.                            */

             sqlo_rc = sqloSSemP(&ipcResources->ipcSemSetID,
                                 &ipcResources->sendSemID,
                                 SQLO_SSEM_UNDO, (U32BIT)SQLO_NO_TIMEOUT); /*@d120677pj*/

             if (sqlo_rc != SQLO_OK)
             {
                sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 12, 8,
                               (void *)&ipcResources->ipcSemSetID, 8,
                               (void *)&ipcResources->sendSemID);
                rc = SQLCC_ERR_COMM_REQUEST_FAILED;
                errors.cleanup = TRUE;                               /*@d16619tjv*/
                ipcseterr(rc,8,"SEMAPHORE WAIT",sqlo_rc);           /*@d14218tjv*/
                goto exit;
             }

             /* Post the synch semaphore to complete initialization */

             sqlo_rc = sqloSSemV(&ipcResources->ipcSemSetID,
                                 &ipcResources->syncSemID, 0);
             if (sqlo_rc != SQLO_OK)
             {
                sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 13, 8,
                               (void *)&ipcResources->ipcSemSetID, 8,
                               (void *)&ipcResources->syncSemID);
                rc = SQLCC_ERR_COMM_REQUEST_FAILED;
                errors.cleanup = TRUE;                               /*@d16619tjv*/
                ipcseterr(rc,8,"POST SEMAPHORE",sqlo_rc);           /*@d14218tjv*/
                goto exit;
             }

             /* don't close the sync semaphore here, otherwise */   /*@d15910SAM*/
             /* on OS/2 it will go away                        */

          }
       }                                                            //@d125700jon
   }
   else
   {                                                             /*@db38558tjv*/
       ipcResources->Allocated = FALSE;
   }                                                             /*@de38558tjv*/

   /* Initialize the connection handle */

   ((SQLCC_IPC_CON_HANDLE_T *)
    pComHandle->connHandle.pConnHandleProto)->aslHeapSz =
                        sqlz_krcbp->sqlekcon.aslheapsz * PAGE4K;

   ((SQLCC_IPC_CON_HANDLE_T *)
    pComHandle->connHandle.pConnHandleProto)->serverFlag =
                                (pInitStruct->server_flag) ? TRUE : FALSE;
   ((SQLCC_IPC_CON_HANDLE_T *)                                     //@d138623swm
    pComHandle->connHandle.pConnHandleProto)->isDariAgent =
                                (pInitStruct->dariRunning) ? TRUE : FALSE;
   sqlomcpy(&((SQLCC_IPC_CON_HANDLE_T *)
              pComHandle->connHandle.pConnHandleProto)->ipcResources,
              ipcResources,
              sizeof(SQLCC_IPC_RESOURCES_T));

   /* Attach to the shared segment for a DARI */                    /*fb12409SAM*/

   if ((pInitStruct->dariFlag)    ||                                /*f12409SAM*/
       (pInitStruct->dariReused))                                   /*f12409SAM*/
   {
     /* Attach to the shared memory communication buffer */

     sqlo_rc = sqlocshr(&ipcResources->ipcShmHandle);
     if (sqlo_rc != SQLO_OK)
     {
        rc = SQLCC_ERR_COMM_REQUEST_FAILED;
        sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 40,
                      sizeof(&ipcResources->ipcShmHandle),
                      (void *)&ipcResources->ipcShmHandle);
        ipcseterr(rc,8,"FREE SHARED MEMORY",sqlo_rc);             /*@d14218tjv*/
        errors.cleanup = TRUE;
        goto exit;
     }
     errors.asl_heap = TRUE;
   }                                                                /*fe12409SAM*/

   /* Initialize the aslCommHeader, at the server and a DARI */

   if ((pInitStruct->dariFlag) ||                                   /*f12409SAM*/
       (pInitStruct->server_flag))
   {
      // only do the calculation if we hasn't done it
      // above. We should already get the aslCommHeader address if we
      // are server and not running dari.
      ((SQLCC_IPC_CON_HANDLE_T *)
       pComHandle->connHandle.pConnHandleProto)->aslCommHeader =
          (aslCommHeader == NULL)?                                  //@db131103xun
          (SQLCC_IPC_ASL_HDR_T *)sqloGetSetRawAddress(&ipcResources->ipcShmHandle):
          aslCommHeader;                                            //@de131103xun

#if defined(NOKILLHANDLER)                                       //@db80418xun
      // if this the server calling ccinit() for dari
      // setup clientPid
      if (pInitStruct->server_flag && pInitStruct->dariRunning)
      {
         ((SQLCC_IPC_CON_HANDLE_T *)
         pComHandle->connHandle.pConnHandleProto)->aslCommHeader->clientPid =
            pGp->pstSqle->s_sqle_agentcb->DariMine->dari_pid;
      }
#endif                                                           //@de80418xun
   }

   /* Extract the current setting of DB2INSTANCE */              /*@db17309tjv*/
   /* and store it in the connection handle.     */
   sqlo_rc = sqlogins(((SQLCC_IPC_CON_HANDLE_T *)
                      pComHandle->connHandle.pConnHandleProto)->instanceName);

   if (sqlo_rc != SQLO_OK)
   {
      rc = SQLCC_ERR_INSUFFICIENT_TARGET_INFO;
   }                                                            /*@de17309tjv*/


exit:

   /* Perform cleanup tasks if errors occured */

   if (errors.cleanup)
   {
      ((SQLCC_IPC_CON_HANDLE_T *)                                 /*@d38558tjv*/
      pComHandle->connHandle.pConnHandleProto)->ipcResources.Allocated = FALSE;

      /* Close the send semaphore if opened */

      if (errors.send_sem)
      {
         sqlo_rc = sqloSSemClose(&ipcResources->ipcSemSetID,
                                 &ipcResources->sendSemID);

      }

      /* Close the recv semaphore if opened */

      if (errors.recv_sem)
      {
         sqlo_rc = sqloSSemClose(&ipcResources->ipcSemSetID,
                                 &ipcResources->recvSemID);

      }

      /* Close the sync semaphore if opened */

      if (errors.sync_sem)                                          /*@d68862xun*/
      {
         sqlo_rc = sqloSSemClose(&ipcResources->ipcSemSetID,
                                 &ipcResources->syncSemID);
      }

      /* Free the protocol specific connection handle */

      if (errors.con_hand)
      {
         sqlo_rc = sqlofmblk(NULL, pComHandle->connHandle.pConnHandleProto);
      }

      if (errors.asl_heap)                                          /*f12409SAM*/
      {                                                             /*f12409SAM*/
         sqlo_rc = sqlodshr(&ipcResources->ipcShmHandle);           /*f12409SAM*/
      }                                                             /*f12409SAM*/

      /* If invoked by the server we must free the shared memory set, */
      /* as well as the individual semaphores.                        */

      // do NOT free the IPC resources if it's agent doing dari       @db68862xun
      // because the dari process or system controler will be
      // responsible to remove the IPC resources
      if ((pInitStruct->server_flag) &&
          !(pInitStruct->dariRunning == TRUE &&
            pInitStruct->dariFlag == FALSE))                        //@de68862xun
      {
         /* the shared memory will be free in sqle_run_agent */    /*d12523ltp*/

         // Set the recvTermFlag in the aslCommHeader to TRUE.      @bd51814tjv
         // If this is not done, when we remove the semaphore
         // set the client code will see the value as FALSE and
         // assume a message was sent from the server.  We
         // will then trap the client as it tries to receive
         // garbage data.
         SQLCC_IPC_ASL_HDR_T * pAslCommHeader =
                  (SQLCC_IPC_ASL_HDR_T *)
                  sqloGetSetRawAddress(&ipcResources->ipcShmHandle);

         pAslCommHeader->recvTermFlag = TRUE;                       //@ed51814tjv

         sqlo_rc = sqloSSemSetFree(&ipcResources->ipcSemSetID,
                                   &ipcResources->recvSemID);
         if (sqlo_rc != SQLO_OK)
         {
            sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 16, 8,
                           (void *)&ipcResources->ipcSemSetID, 8,
                           (void *)&ipcResources->recvSemID);

            sqlt_logerr_zrc (SQLT_SQLCC, SQLT_SQLCCIPCINIT, 3,      /*@d18956glh*/
                             sqlo_rc,
                             SQLT_FFSL_SEV | SQLT_ADMIN);/*@d106920jh*/ /*@d18956glh*/

            sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,     /*@d18956glh*/
                             (void *) &ipcResources->ipcSemSetID,   /*@d18956glh*/
                             8, SQLT_FFSL_SEV);                     /*@d18956glh*/

            sqlt_logerr_data((char *) SQLCC_FFDC_RECV_SEMID,        /*@d18956glh*/
                             (void *) &ipcResources->recvSemID,     /*@d18956glh*/
                             8, SQLT_FFSL_SEV);                     /*@d18956glh*/

            rc = SQLCC_ERR_SYSTEM_ERROR;
         }

         sqlo_rc = sqloSSemSetFree(&ipcResources->ipcSemSetID,
                                   &ipcResources->sendSemID);
         if (sqlo_rc != SQLO_OK)
         {
            sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 17, 8,
                           (void *)&ipcResources->ipcSemSetID, 8,
                           (void *)&ipcResources->sendSemID);

            sqlt_logerr_zrc (SQLT_SQLCC, SQLT_SQLCCIPCINIT, 6,      /*@d18956glh*/
                             sqlo_rc,
                             SQLT_FFSL_SEV | SQLT_ADMIN);/*@d106920jh*/  /*@d18956glh*/

            sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,     /*@d18956glh*/
                             (void *) &ipcResources->ipcSemSetID,   /*@d18956glh*/
                             8, SQLT_FFSL_SEV);                     /*@d18956glh*/

            sqlt_logerr_data((char *) SQLCC_FFDC_SEND_SEMID,        /*@d18956glh*/
                             (void *) &ipcResources->sendSemID,     /*@d18956glh*/
                             8, SQLT_FFSL_SEV);                     /*@d18956glh*/

            rc = SQLCC_ERR_SYSTEM_ERROR;
         }

         if (errors.freeSync)                                    /*@db22959tjv*/
         {
            sqlo_rc = sqloSSemSetFree(&ipcResources->ipcSemSetID,
                                      &ipcResources->syncSemID);
            if (sqlo_rc != SQLO_OK)
            {
               sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 18, 8,
                              (void *)&ipcResources->ipcSemSetID, 8,
                              (void *)&ipcResources->syncSemID);

               sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 9,     /*@d18956glh*/
                               sqlo_rc,
                               SQLT_FFSL_SEV | SQLT_ADMIN);/*@d106920jh*/ /*@d18956glh*/

               sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,   /*@d18956glh*/
                                (void *) &ipcResources->ipcSemSetID, /*@d18956glh*/
                                8, SQLT_FFSL_SEV);                   /*@d18956glh*/

               sqlt_logerr_data((char *) SQLCC_FFDC_SYNC_SEMID,      /*@d18956glh*/
                                (void *) &ipcResources->syncSemID,   /*@d18956glh*/
                                8, SQLT_FFSL_SEV);                   /*@d18956glh*/

               rc = SQLCC_ERR_SYSTEM_ERROR;
            }

            // Free the semaphore set                                 @bd51814tjv
            sqlo_rc = sqloSSemSetRemove(&ipcResources->ipcSemSetID);

            if (sqlo_rc != SQLO_OK)
            {
               sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 19,
                             sizeof(ipcResources->ipcSemSetID),
                             (void *)&ipcResources->ipcSemSetID);

               sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCIPCINIT, 10,
                               sqlo_rc,
                               SQLT_FFSL_SEV | SQLT_ADMIN);      /*@d106920jh*/

               sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,
                                (void *) &ipcResources->ipcSemSetID,
                                8, SQLT_FFSL_SEV);

               rc = SQLCC_ERR_SYSTEM_ERROR;
            }                                                    //@ed51814tjv
         }                                                       /*@de22959tjv*/
      }
   }
   else if ((pInitStruct->dariFlag) ||                           /*@db38558tjv*/
            (pInitStruct->server_flag))
   {
      ((SQLCC_IPC_CON_HANDLE_T *)
      pComHandle->connHandle.pConnHandleProto)->ipcResources.Allocated = TRUE;
   }                                                             /*@de38558tjv*/

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCINIT, (int) rc);
   return(rc);
}
#endif //BLD_CLIENT_ONLY

/*****************************************************************************
*
*  Function Name     = sqlccipcdarihandshake
*
*  Descriptive Name  = Handshake the semaphore initialization with a server
*
*  Function          = This function is used by a DARI process that has
*                      already done a CCINIT, but needs to handshake the
*                      posting and setting of the IPC semaphores, for a
*                      new server process (agent) that is just attaching to the
*                      IPC resources, and is doing a CCINIT.
*
*  Input             = 1) initStruct: a structure containing a pointer
*                                     to the Connection Handle.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_COMM_REQUEST_FAILED
*
*****************************************************************************/
                                                                    /*@fb12409SAM*/
short sqlccipcdarihandshake(
                   SQLCC_INITSTRUCT_T            *pInitStruct
                  )
{
   int                     rc;
   SQLCC_IPC_RESOURCES_T  *ipcResources;


   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCDARIHANDSHAKE);

   ipcResources = (SQLCC_IPC_RESOURCES_T *)&pInitStruct->ipcResources;

   /* Wait on the server's receive semaphore, this will be posted  */
   /* by the server.  This operation is performed to ensure that   */
   /* the dari process performs an UNDO operation on the server's  */
   /* receive semaphore.  This will ensure that the server is      */
   /* posted if the dari process exits.                            */

   rc = sqloSSemP(&ipcResources->ipcSemSetID, &ipcResources->sendSemID,
                  SQLO_SSEM_INTERRUPTIBLE, (U32BIT)SQLO_NO_TIMEOUT); //@d61259tjv /*@d120677pj*/

   if (rc != SQLO_OK)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCDARIHANDSHAKE, 1, 8,
                     (void *)&ipcResources->ipcSemSetID, 8,
                     (void *)&ipcResources->sendSemID);
      sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCIPCDARIHANDSHAKE, 3,    //@d46356tjv
                      rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/ //@d46356tjv
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
      goto exit;
   }

   /* The synch sem should already be opened by the DARI, and we */ /*@d15910SAM*/
   /* don't want to close it until we do a sqlccdariterm()       */ /*@d15910SAM*/

   /* Post the synch semaphore to complete initialization */
   rc = sqloSSemV(&ipcResources->ipcSemSetID, &ipcResources->syncSemID, 0);
   if (rc != SQLO_OK)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCDARIHANDSHAKE, 3, 8,
                     (void *)&ipcResources->ipcSemSetID, 8,
                     (void *)&ipcResources->syncSemID);
      sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCIPCDARIHANDSHAKE, 4,    //@d46356tjv
                      rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/   //@d46356tjv
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
   }

exit:

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCDARIHANDSHAKE, (int) rc);
   return(rc);
}
                                                                    /*@fe12409SAM*/

#ifndef BLD_CLIENT_ONLY

/*****************************************************************************
*
*  Function Name     = sqlccipcconnr
*
*  Component Name    = SQLC
*
*  Descriptive Name  = IPC protocol connect to server
*
*  Func. Prototype   = short sqlccipcconnr(
*                             SQLCC_CONNECT_T       *pConnect,
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             SQLCC_COND_T          *pCondition
*                                       )
*
*  Function          = This function uses the directory information in the
*                      node directory to determine the instance which
*                      contains the database.  If there was no entry in
*                      the node directory the instane name will be extracted
*                      from the DB2INSTANCE environment variable.  The
*                      instance name will be used to construct a unique key
*                      which will be used to connect to the IPC listener's
*                      queue.
*
*  Input             = 1) pNodeDir:   a pointer to an SQLE_NECB type
*                                     structure that contains the
*                                     target information required
*                                     to connect to a DB2 Instance
*
*                      2) pComHandle: a pointer to the Communication
*                                     Handle that contains pointers
*                                     to all the IPC specific
*                                     functions and data structures.
*
*  Output            = 1) pCondition: a pointer to the common Condition
*                                     structure that contains a pointer
*                                     to the IPC specific condition
*                                     structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_CONN_CLOSED_BY_PARTNER             @d14218tjv
*                      SQLCC_ERR_INSUFFICIENT_TARGET_INFO
*                      SQLCC_ERR_NO_DB2START                        @d11917tjv
*
*****************************************************************************/

short sqlccipcconnr(
                    SQLCC_CONNECT_T               *pConnect,
                    SQLCC_COMHANDLE_T             *pComHandle,
                    SQLCC_COND_T                  *pCondition
                   )
{
   short                    rc      = SQLCC_OK;
   int                      sqlo_rc = SQLO_OK;
   struct sqlo_pidinfo      process_ids_area;
   SQLCC_IPC_CON_REQST_T    request_msg;
   SQLCC_IPC_CON_REPLY_T    reply_msg;
   SQLER_GLOB * erglob_ptr;                                       /*@d63024kcs*/

   struct  /* Flags to assist in error handling */
   {
      unsigned  cleanup  : 1;
      unsigned  send_sem : 1;
      unsigned  recv_sem : 1;
      unsigned  sync_sem : 1;
      unsigned  asl_heap : 1;
   } errors = {0};

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCCONNR);

   struct sql_static_data * pGp = sqlo_get_static_data();         /*@d63024kcs*/

   erglob_ptr = &pGp->pstSqleUbi->s_sqlerglob;                    /*@d63024kcs*/

   /* Determine the pid of the client */

   sqlo_rc = sqlogpid(&process_ids_area);
   if (sqlo_rc != SQLO_OK)
   {
     rc = SQLCC_ERR_SYSTEM_ERROR;
     goto exit;
   }

   /* If a instancename was specified in the NECB reset */
   /* the instanceName to this value.                   */
   if (pConnect->pNodeDir->instance_name[0] != '\0')
   {
      sqlomcpy(&((SQLCC_IPC_CON_HANDLE_T *)                      /*@db16619tjv*/
               pComHandle->connHandle.pConnHandleProto)->instanceName[0],
               pConnect->pNodeDir->instance_name,
               sizeof(pConnect->pNodeDir->nodename));
   }                                                             /*@db16619tjv*/

   /* Set up the request message, and the reply message header */
   request_msg.msg_hdr.msg_buf_size = sizeof(request_msg)-                     /*@d13112mah*/
                                      sizeof(request_msg.msg_hdr.msg_buf_size);/*@d13112mah*/
   request_msg.type                 = SQLCC_IPC_CONNECT;
   request_msg.client_pid           = process_ids_area.procid_cpid;
   request_msg.client_tid           = process_ids_area.procid_ctid;
   request_msg.eff_uid              = sqlo_get_uid();

   reply_msg.msg_hdr.msg_buf_size   = sizeof(reply_msg)-                       /*@d13112mah*/
                                      sizeof(reply_msg.msg_hdr.msg_buf_size);  /*@d13112mah*/

   /* Send connection request to the IPC listener */
   /* and receive the associated reply.           */

   sqlo_rc = sqloTQMRequestMLN(
                    ((SQLCC_IPC_CON_HANDLE_T *)
                    pComHandle->connHandle.pConnHandleProto)->instanceName,
                    SQLO_QUE_IPCLIST,
                    (SQLO_MSG_HDR *)&request_msg,
                    SQLOQ_HIGH,
                    (SQLO_MSG_HDR *)&reply_msg,
                    SQLO_QUE_WAIT | SQLO_QUE_NO_TRUN,
                    SQLO_NO_TIMEOUT,
                    sqlz_krcbp->sqlekcon.nodetype,                /*@d44244jkl*/
                    erglob_ptr->lConnAttNode);                    /*@d63024kcs*/

   sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCIPCCONNR, 100,
                 sizeof(SQLCC_IPC_RESOURCES_T),
                 &reply_msg.ipcResources);

   if (sqlo_rc != SQLO_OK)
   {
      rc = SQLCC_ERR_NO_DB2START;                                 /*@d11917tjv*/
      ipcseterr(rc,8,"QUEUE WRITE",sqlo_rc);
      goto exit;
   }
   else if (reply_msg.rc != SQLCC_OK)
   {
      rc = reply_msg.rc;                                          /*@d11917tjv*/
      ipcseterr(rc,8,reply_msg.token,reply_msg.rc);               /*@d14218tjv*/
      goto exit;
   }

   /* Reset the ASLHEAPSZ value to that of the server.  This */  /*@db14108tjv*/
   /* ensures that we synchronize the client's size with     */
   /* that if the server.  These values can get out of synch */
   /* if an application does not exit after the database     */
   /* manager configuration value is updated and the server  */
   /* restarted.                                             */
   ((SQLCC_IPC_CON_HANDLE_T *)
    pComHandle->connHandle.pConnHandleProto)->aslHeapSz = reply_msg.aslHeapSz;

                                                                 /*@de14108tjv*/

   /* Open the send semaphore and then post it.  This task is required */
   /* to ensure that both the client and server initialize the send    */
   /* receive semaphores properly.  The server will open its receive   */
   /* semaphore first and wait on this post.                           */

   sqlo_rc = sqloSSemOpen(&reply_msg.ipcResources.ipcSemSetID,
                          &reply_msg.ipcResources.sendSemID);
   if (sqlo_rc != SQLO_OK)
   {
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                      /*@d14218tjv*/
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCCONNR, 1, 8,
                     (void *)&reply_msg.ipcResources.ipcSemSetID, 8,
                     (void *)&reply_msg.ipcResources.sendSemID);
      ipcseterr(rc,8,"OPEN SEMAPHORE",sqlo_rc);                   /*@d14218tjv*/
      errors.cleanup = TRUE;
      goto exit;
   }
   errors.send_sem = TRUE;

   /* Open the receive semaphore */

   sqlo_rc = sqloSSemOpen(&reply_msg.ipcResources.ipcSemSetID,
                          &reply_msg.ipcResources.recvSemID);
   if (sqlo_rc != SQLO_OK)
   {
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                      /*@d14218tjv*/
      ipcseterr(rc,8,"OPEN SEMAPHORE",sqlo_rc);                   /*@d14218tjv*/
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCCONNR, 2, 8,
                     (void *)&reply_msg.ipcResources.ipcSemSetID, 8,
                     (void *)&reply_msg.ipcResources.recvSemID);

      errors.cleanup = TRUE;
      goto exit;
   }
   errors.recv_sem = TRUE;

   /* Open the sync semaphore */

   sqlo_rc = sqloSSemOpen(&reply_msg.ipcResources.ipcSemSetID,
                          &reply_msg.ipcResources.syncSemID);
   if (sqlo_rc != SQLO_OK)
   {
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                      /*@d14218tjv*/
      ipcseterr(rc,8,"OPEN SEMAPHORE",sqlo_rc);                   /*@d14218tjv*/
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCCONNR, 3, 8,
                     (void *)&reply_msg.ipcResources.ipcSemSetID, 8,
                     (void *)&reply_msg.ipcResources.syncSemID);

      errors.cleanup = TRUE;
      goto exit;
   }
   errors.sync_sem = TRUE;

   /* Attach to the shared memory communication buffer */

   sqlo_rc = sqlocshr2(&reply_msg.ipcResources.ipcShmHandle,       /*d35497aha*/
                       ((SQLCC_IPC_CON_HANDLE_T *)                 /*d35497aha*/
                       pComHandle->connHandle.pConnHandleProto)->instanceName,
                       SQLT_FFSL_SEV);                             //@d134598exo
   if (sqlo_rc != SQLO_OK)
   {
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                      /*@d14218tjv*/
      ipcseterr(rc,8,"FREE SHARED MEMORY",sqlo_rc);               /*@d14218tjv*/
      sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCIPCCONNR, 4,
                    sizeof(reply_msg.ipcResources.ipcShmHandle),
                    (void *)&reply_msg.ipcResources.ipcShmHandle);

      errors.cleanup = TRUE;
      goto exit;
   }
   errors.asl_heap = TRUE;

   /* Anchor the shared memory address in the communication handle *          */

   ((SQLCC_IPC_CON_HANDLE_T *)
     pComHandle->connHandle.pConnHandleProto)->aslCommHeader =
                    (SQLCC_IPC_ASL_HDR_T *)
                    sqloGetSetRawAddress(&reply_msg.ipcResources.ipcShmHandle);

   #if defined(NOKILLHANDLER)                                       //@bd80241tjv

   // Set the client's PID in the aslCommHeader.  This will be used by
   // the server to periodically check whether the client has disappeared.

   // Move the setting of clientPid to sqlccipcconmgr_process_conn   @d154946ltp

   // ((SQLCC_IPC_CON_HANDLE_T *)
   //  pComHandle->connHandle.pConnHandleProto)->aslCommHeader->clientPid =
   //                                               process_ids_area.procid_cpid;

   #endif                                                           //@ed80241tjv

   /* Copy the IPC resources into the communication handle. The client's */
   /* send semaphore is the server's receive semaphore and vice versa.   */

   sqlomcpy(&((SQLCC_IPC_CON_HANDLE_T *)
        pComHandle->connHandle.pConnHandleProto)->ipcResources.ipcShmHandle,
        &reply_msg.ipcResources.ipcShmHandle,
        sizeof(SQLO_MEM_SET));

   ((SQLCC_IPC_CON_HANDLE_T *)
    pComHandle->connHandle.pConnHandleProto)->ipcResources.ipcShmSize =
                                          reply_msg.ipcResources.ipcShmSize;

   sqlomcpy(&((SQLCC_IPC_CON_HANDLE_T *)
        pComHandle->connHandle.pConnHandleProto)->ipcResources.ipcSemSetID,
        &reply_msg.ipcResources.ipcSemSetID,
        sizeof(SQLO_SSEMSETID));

   sqlomcpy(&((SQLCC_IPC_CON_HANDLE_T *)
        pComHandle->connHandle.pConnHandleProto)->ipcResources.recvSemID,
        &reply_msg.ipcResources.sendSemID,
        sizeof(SQLO_SSEMID));

   sqlomcpy(&((SQLCC_IPC_CON_HANDLE_T *)
        pComHandle->connHandle.pConnHandleProto)->ipcResources.sendSemID,
        &reply_msg.ipcResources.recvSemID,
        sizeof(SQLO_SSEMID));

   /* Wait on the server's receive semaphore, this will be posted by the */
   /* server.  This operation is performed to ensure we synchronize our  */
   /* startup, and to ensure we have an UNDO operation on the server's   */
   /* receive semaphore.  The UNDO will ensure that the server is posted */
   /* if the application process exits.                                  */

   sqlo_rc = sqloSSemP(&reply_msg.ipcResources.ipcSemSetID,
                       &reply_msg.ipcResources.recvSemID,
                       SQLO_SSEM_UNDO, (U32BIT)SQLO_NO_TIMEOUT);  /*@d120677pj*/

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCCONNR, 5, 8,
                     (void *)&reply_msg.ipcResources.ipcSemSetID, 8,
                     (void *)&reply_msg.ipcResources.recvSemID);
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                      /*@d14218tjv*/
      errors.cleanup = TRUE;                                      /*@d16619tjv*/
      ipcseterr(rc,8,"SEMAPHORE WAIT",sqlo_rc);                   /*@d14218tjv*/
      goto exit;
   }

   /* Post the synch semaphore to complete initialization */

   sqlo_rc = sqloSSemV(&reply_msg.ipcResources.ipcSemSetID,
                       &reply_msg.ipcResources.syncSemID, 0);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCCONNR, 7, 8,
                     (void *)&reply_msg.ipcResources.ipcSemSetID, 8,
                     (void *)&reply_msg.ipcResources.syncSemID);
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                      /*@d14218tjv*/
      errors.cleanup = TRUE;                                      /*@d16619tjv*/
      ipcseterr(rc,8,"SEMAPHORE POST",sqlo_rc);                   /*@d14218tjv*/
      goto exit;
   }

   /* Close the sync semaphore */

   sqlo_rc = sqloSSemClose(&reply_msg.ipcResources.ipcSemSetID,
                           &reply_msg.ipcResources.syncSemID);

   errors.sync_sem = FALSE;
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCCONNR, 8, 8,
                     (void *)&reply_msg.ipcResources.ipcSemSetID, 8,
                     (void *)&reply_msg.ipcResources.syncSemID);
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                      /*@d14218tjv*/
      ipcseterr(rc,8,"SEMAPHORE CLOSE",sqlo_rc);                  /*@d14218tjv*/
      errors.cleanup = TRUE;
      goto exit;
   }

exit:

   /* Perform cleanup tasks if errors occured */

   if (errors.cleanup)
   {
      /* Post and close the sync semaphore if opened */          /*@db16619tjv*/
      if (errors.sync_sem)
      {
         sqlo_rc = sqloSSemV(&reply_msg.ipcResources.ipcSemSetID,
                             &reply_msg.ipcResources.syncSemID, 0);

         sqlo_rc = sqloSSemClose(&reply_msg.ipcResources.ipcSemSetID,
                                 &reply_msg.ipcResources.syncSemID);
      }                                                          /*@de16619tjv*/

      /* Post and close the send semaphore if opened */

      if (errors.send_sem)
      {
         sqlo_rc = sqloSSemV(&reply_msg.ipcResources.ipcSemSetID,
                             &reply_msg.ipcResources.sendSemID, 0);

         sqlo_rc = sqloSSemClose(&reply_msg.ipcResources.ipcSemSetID,
                                 &reply_msg.ipcResources.sendSemID);
      }

      /* Close the recv semaphore if opened */

      if (errors.recv_sem)
      {
         sqlo_rc = sqloSSemClose(&reply_msg.ipcResources.ipcSemSetID,
                                 &reply_msg.ipcResources.recvSemID);

      }

      /* Detach from the shared memory communication buffer */

      if (errors.asl_heap)
      {
         sqlo_rc = sqlodshr(&reply_msg.ipcResources.ipcShmHandle);
      }
   }
   else                                                          /*@db38558tjv*/
   {
      ((SQLCC_IPC_CON_HANDLE_T *)
      pComHandle->connHandle.pConnHandleProto)->ipcResources.Allocated = TRUE;
   }                                                             /*@de38558tjv*/

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCCONNR, (int) rc);
   return(rc);
}

/*****************************************************************************
*
*  Function Name     = sqlccipcgetbuffer
*
*  Component Name    = SQLCC
*
*  Descriptive Name  = IPC protocol support communication buffer allocation
*
*  Func. Prototype   = short sqlccipcgetbuffer(
*                                                                     @fb12116pme
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             SQLCC_COND_T          *pCondition
*                                                                     @fe12116pme
*                                      )
*
*  Function          = This function allocates a communication buffer of the
*                      specified size.  For IPC communications this size
*                      must match the size of the ASL heap.  The first
*                      call to this function will simply return a pointer
*                      to the shared communications buffer.  Subsequent calls
*                      will allocate a buffer from private memory.
*
*                      The private buffer will be used primarily for blocking
*                      cursors where the cache of rows is retained in the
*                      communications buffer.
*
*  Input             = 1) pComHandle : A pointer to the Communication
*                                      Handle that contains pointers
*                                      to all the IPC specific
*                                      functions and data structures.
*
*
*  Output              1) pCondition : A pointer to the common Condition
*                                      structure that contains a pointer
*                                      to the IPC specific condition
*                                      structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_MEM_UNAVAIL
*                      SQLCC_ERR_INV_DATA_LENGTH
*
*****************************************************************************/

                                                                    /*@fb12116pme*/
short sqlccipcgetbuffer (SQLCC_COMHANDLE_T  * pComHandle,
                         SQLCC_COND_T       * pCondition)
                                                                    /*@fe12116pme*/
{
   short   rc = SQLCC_OK;
   int     sqlo_rc = SQLO_OK;

   U32BIT        bufferSize;                                        /*@f12116pme*/
   char       ** pbuffer;                                           /*@f12116pme*/
   U32BIT        aslSize;                                           /*@d61766kcs*/

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCGETBUFFER);

   bufferSize = pComHandle->comGetBuffer.ulReqBuffLen;              /*@f12116pme*/
   pbuffer    = pComHandle->comGetBuffer.ppcBuffAddr;               /*@f12116pme*/

   /* Ensure that the size request matches the ASL heap size */


   if (bufferSize > (((SQLCC_IPC_CON_HANDLE_T *)                  /*@d61766kcs*/
                       pComHandle->connHandle.pConnHandleProto)->aslHeapSz))
   {
      sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCIPCGETBUFFER, 1,
                       sizeof(bufferSize), (void *)&bufferSize);

      aslSize = (((SQLCC_IPC_CON_HANDLE_T *)                      /*@d61766kcs*/
                  pComHandle->connHandle.pConnHandleProto)->aslHeapSz);/*@d61766kcs*/
      sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCIPCGETBUFFER, 10,       /*@d61766kcs*/
                  4, &aslSize);                                   /*@d61766kcs*/

      rc = SQLCC_ERR_INV_DATA_LENGTH;
      goto exit;
   }

   /* If this is the first call to this routine simply return */
   /* a pointer to the shared communication buffer.           */

   if (((SQLCC_IPC_CON_HANDLE_T *)
        pComHandle->connHandle.pConnHandleProto)->commBuffer == NULL)
   {
      *pbuffer = (char *)((SQLCC_IPC_CON_HANDLE_T *)
                  pComHandle->connHandle.pConnHandleProto)->aslCommHeader +
                  sizeof(SQLCC_IPC_ASL_HDR_T);

      ((SQLCC_IPC_CON_HANDLE_T *)
       pComHandle->connHandle.pConnHandleProto)->commBuffer = *pbuffer;
   }
   else
   {

      /* Allocate a communication buffer from private memory */

      sqlo_rc = sqlogblk(NULL,
                         ((SQLCC_IPC_CON_HANDLE_T *)
                         pComHandle->connHandle.pConnHandleProto)->aslHeapSz,
                         (void **)pbuffer);

      if (sqlo_rc != SQLO_OK )
      {
         sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCIPCGETBUFFER, 2,
                          sizeof(sqlo_rc), (void *) &sqlo_rc);
         rc = SQLCC_ERR_MEM_UNAVAIL;
      }
   }

exit:

                                                                    /*@fb12116pme*/
   if (rc == SQLCC_OK) {
     pComHandle->comGetBuffer.ulRetBuffLen = bufferSize;
   } else {
     pComHandle->comGetBuffer.ulRetBuffLen = 0;
   } /* endif */
                                                                    /*@fe12116pme*/

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCGETBUFFER, (int) rc);
   return(rc);
}

/*****************************************************************************
*
*  Function Name     = sqlccipcfreebuffer
*
*  Component Name    = SQLCC
*
*  Descriptive Name  = IPC protocol support communication buffer deallocation
*
*  Func. Prototype   = short sqlccipcfreebuffer(
*                                                                     @fb12116pme
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             SQLCC_COND_T          *pCondition
*                                                                     @fe12116pme
*                                              )
*
*  Function          = This function deallocates the specified communication
*                      buffer.  If the buffer address matches the shared
*                      communication buffer, the buffer pointer in the
*                      connection handle is simply set to NULL.
*
*  Input             = 1) pComHandle : A pointer to the Communication
*                                      Handle that contains pointers
*                                      to all the IPC specific
*                                      functions and data structures.
*                                      All these data structures
*                                      are freed by this function.
*
*
*  Output            = 1) pCondition : A pointer to the common Condition
*                                      structure that contains a pointer
*                                      to the IPC specific condition
*                                      structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_MEM_NOT_FREED
*
*****************************************************************************/

                                                                    /*@fb12116pme*/
short sqlccipcfreebuffer ( SQLCC_COMHANDLE_T      *pComHandle,
                           SQLCC_COND_T           *pCondition )
                                                                    /*@fe12116pme*/

{
   short     rc      = SQLCC_OK;
   int       sqlo_rc = SQLO_OK;

   char        * pbuffer;                                           /*@f12116pme*/

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCFREEBUFFER);

   pbuffer    = pComHandle->comFreeBuffer.pcBuffAddr;               /*@f12116pme*/

   /* If the buffer matches the shared communication buffer */
   /* set the shared buffer comm buffer pointer to NULL     */

   if (pbuffer == ((SQLCC_IPC_CON_HANDLE_T *)
                   pComHandle->connHandle.pConnHandleProto)->commBuffer)
   {
      ((SQLCC_IPC_CON_HANDLE_T *)
       pComHandle->connHandle.pConnHandleProto)->commBuffer = NULL;
       pComHandle->comFreeBuffer.pcBuffAddr = NULL;               /*@d14108tkv*/
   }
   else
   {
      if (pbuffer != NULL)                                        /*@d14108tjv*/
      {
         /* The buffer was allocated out of private memory, free it */

         sqlo_rc = sqlofmblk(NULL, pbuffer);

         if (sqlo_rc != SQLO_OK)
         {
            rc = SQLCC_ERR_MEM_NOT_FREED;
         }
      }                                                           /*@d14108tjv*/
   }

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCFREEBUFFER, (int) rc);
   return(rc);
}

/*****************************************************************************
*
*  Function Name     = sqlccipcdisc
*
*  Component Name    = SQLCC
*
*  Descriptive Name  = IPC protocol close connection
*
*  Func. Prototype   = short sqlccipcdisc(
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             SQLCC_COND_T          *pCondition
*                                      )
*
*  Function          = This function disconnects the process from its peer
*                      IPC process.  There are two forms of disconnect;
*                      NORMAL and ABNORMAL.  If an ABNORMAL disconnect has
*                      been requested the peer IPC process will return an
*                      abnormal condition when it invokes the next IPC call.
*
*  Input             = 1) pComHandle: a pointer to the Communication
*                                     Handle that contains pointers
*                                     to all the IPC specific
*                                     functions and data structures.
*
*  Output            = 1) pCondition: a pointer to the common Condition
*                                     structure that contains a pointer
*                                     to the IPC specific condition
*                                     structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_CONN_CLOSED_BY_PARTNER             @d14218tjv
*                      SQLCC_ERR_NO_CONNECTION
*
*****************************************************************************/

short sqlccipcdisc ( SQLCC_COMHANDLE_T       *pComHandle,
                     SQLCC_COND_T            *pCondition )

{
   short                    rc      = SQLCC_OK;
   int                      sqlo_rc = SQLO_OK;
   SQLCC_IPC_CON_HANDLE_T * ipcConHandle;

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCDISC);

   /* Ensure that we have a connection */

   if (pComHandle->connHandle.pConnHandleProto == NULL)
   {
      rc = SQLCC_ERR_NO_CONNECTION;
      goto exit;
   }

   /* Set up a local connection handle for readability */

   ipcConHandle = (SQLCC_IPC_CON_HANDLE_T *)
                  pComHandle->connHandle.pConnHandleProto;

   /* If we are executing on the server do nothing, the disconnect */
   /* will always be driven by the client application.             */

   if (ipcConHandle->serverFlag == TRUE)
   {

      goto exit;
   }

   /* If this is a NORMAL disconnect request set the disconnect      */
   /* termination flags on, and set the receive termination flag off */

   if (pComHandle->comDisc.ucDiscType == SQLCC_DISCONNECT_NORMAL)
   {
      ipcConHandle->aslCommHeader->disconnectRqst = TRUE;
      ipcConHandle->aslCommHeader->recvTermFlag   = FALSE;
   }
   else
   {
      /* ABNORMAL termination request set the receive termination */
      /* flag on, and turn the disconnect flag off.               */

      ipcConHandle->aslCommHeader->disconnectRqst = FALSE;
      ipcConHandle->aslCommHeader->recvTermFlag   = TRUE;
   }

   /* Post the receive semaphore to wake up the corresponding process */

   sqlo_rc = sqloSSemV(&ipcConHandle->ipcResources.ipcSemSetID,
                       &ipcConHandle->ipcResources.sendSemID, 0);

   if (sqlo_rc != SQLO_OK)
   {
                                                                 /*@bd46979scr*/
      if (pComHandle->comDisc.ucDiscType == SQLCC_DISCONNECT_ABNORMAL)
      {
        sqlt_logerr_data("Abnormal connect termination; semV err OK",
                "",0,SQLT_FFSL_INF);                               /*@d89989pj*/
        sqlo_rc = SQLO_OK;
      }
      else
      {
                                                                 /*@ed46979scr*/
        sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCDISC, 1, 8,
                       (void *)&ipcConHandle->ipcResources.ipcSemSetID, 8,
                       (void *)&ipcConHandle->ipcResources.sendSemID);
        rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                      /*@d14218tjv*/
        ipcseterr(rc,8,"SEMAPHORE POST",sqlo_rc);                   /*@d14218tjv*/
        goto exit;
      }                                                           /*@d46979scr*/
   }

exit:

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCDISC, (int) rc);
   return(rc);
}

/*****************************************************************************
*
*  Function Name     = sqlccipcterm
*
*  Component Name    = SQLCC
*
*  Descriptive Name  = IPC protocol support termination
*
*  Func. Prototype   = short sqlccipcterm(
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             SQLCC_COND_T          *pCondition
*                                      )
*
*  Function          = This function performs resource clean-up task
*                      for a disconnected IPC connection.  It frees
*                      up all IPC specific data structures acquired
*                      during the sqlccipcinit processing, and also
*                      disconnects the process from the IPC resources.
*
*  Input             = 1) pComHandle: a pointer to the Communication
*                                     Handle that contains pointers
*                                     to all the IPC specific
*                                     functions and data structures.
*                                     All these data structures
*                                     are freed by this function.
*
*  Output            = 1) pCondition: a pointer to the common Condition
*                                     structure that contains a pointer
*                                     to the IPC specific condition
*                                     structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_MEM_NOT_FREED
*                      SQLCC_ERR_SYSTEM_ERROR
*                      SQLCC_ERR_NO_CONNECTION
*
*****************************************************************************/

short sqlccipcterm ( SQLCC_COMHANDLE_T   * pComHandle,
                     SQLCC_COND_T        * pCondition )

{
   short        rc      = SQLCC_OK;
   int          sqlo_rc = SQLO_OK;
   SQLCC_IPC_CON_HANDLE_T * ipcConHandle;

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCTERM);

   /* Ensure that we have a connection */

   if (pComHandle->connHandle.pConnHandleProto == NULL)
   {
      rc = SQLCC_ERR_NO_CONNECTION;
      goto exit;
   }

   /* Set up a local connection handle for readability */

   ipcConHandle = (SQLCC_IPC_CON_HANDLE_T *)
                  pComHandle->connHandle.pConnHandleProto;


   /* To be sure that the client finds out about the FORCEing      @bd13843mah*/
   /* of an agent, we need to post the server's send semaphore.  Call         */
   /* the exit list routine to do this.  Then de-install the exit list        */
   /* routine (if we leave it installed, it'll trap as it tries to access     */
   /* freed memory when the EDU normally exits).                              */
   if (ipcConHandle->serverFlag == TRUE)
   {
      sqlccipcAgentExitList ( 1 /* dummy value */ );

      sqlo_rc = sqloRemoveEDUExitList (
#if defined(SQLWINT_VA)                    /*@d104548gal*/      /*@d74674gal*/
                                        (SQLO_EDU_EXIT*)        /*@d74674gal*/
#endif                                                          /*@d74674gal*/
                                        sqlccipcAgentExitList );
      if (sqlo_rc != SQLO_OK)
      {
         sqlt_system_error (SQLT_SQLCC, SQLT_SQLCCIPCTERM, 2, sizeof(sqlo_rc), &sqlo_rc);
         rc = SQLCC_ERR_SYSTEM_ERROR;
      }                                                          /*@ed13843mah*/
   }

   if (ipcConHandle->ipcResources.Allocated)                     /*@db38558tjv*/
   {
      /* Close the receive semaphore, if the close fails set the */
      /* return code, issue a trace, and continue.               */

      sqlo_rc = sqloSSemClose(&ipcConHandle->ipcResources.ipcSemSetID,
                              &ipcConHandle->ipcResources.recvSemID);

      if (sqlo_rc != SQLO_OK)
      {
         sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCTERM, 4, 8,
                        (void *)&ipcConHandle->ipcResources.ipcSemSetID,
                        8, (void *)&ipcConHandle->ipcResources.recvSemID);

         rc = SQLCC_ERR_SYSTEM_ERROR;
         ipcseterr(rc,8,"SEMAPHORE CLOSE",sqlo_rc);                  /*@d14218tjv*/
      }

      /* Close the send semaphore, if the close fails set the */
      /* return code, issue a trace, and continue.            */

      sqlo_rc = sqloSSemClose(&ipcConHandle->ipcResources.ipcSemSetID,
                              &ipcConHandle->ipcResources.sendSemID);

      if (sqlo_rc != SQLO_OK)
      {
         sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCTERM, 6, 8,
                        (void *)&ipcConHandle->ipcResources.ipcSemSetID,
                        8, (void *)&ipcConHandle->ipcResources.sendSemID);

         rc = SQLCC_ERR_SYSTEM_ERROR;
         ipcseterr(rc,8,"SEMAPHORE CLOSE",sqlo_rc);                  /*@d14218tjv*/
      }

   }

   /* If we are executing on the server free the shared memory set. */
   /* otherwise just disconnect from it.                            */

   if (ipcConHandle->serverFlag == TRUE)
   {
      /* the shared memory will be free in sqle_run_agent */       /*d12523ltp*/

      /* Free the send semaphore back to the system */

      sqlo_rc = sqloSSemSetFree(&ipcConHandle->ipcResources.ipcSemSetID,
                                &ipcConHandle->ipcResources.sendSemID);

      if (sqlo_rc != SQLO_OK)
      {
         sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCTERM, 10, 8,
                        (void *)&ipcConHandle->ipcResources.ipcSemSetID,
                        8, (void *)&ipcConHandle->ipcResources.sendSemID);

         rc = SQLCC_ERR_SYSTEM_ERROR;
      ipcseterr(rc,8,"SEMAPHORE FREE",sqlo_rc);                   /*@d14218tjv*/
      }

      /* Free the receive semaphore back to the system */

      sqlo_rc = sqloSSemSetFree(&ipcConHandle->ipcResources.ipcSemSetID,
                                &ipcConHandle->ipcResources.recvSemID);

      if (sqlo_rc != SQLO_OK)
      {
         sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCTERM, 12, 8,
                        (void *)&ipcConHandle->ipcResources.ipcSemSetID,
                        8, (void *)&ipcConHandle->ipcResources.recvSemID);

         rc = SQLCC_ERR_SYSTEM_ERROR;
         ipcseterr(rc,8,"SEMAPHORE FREE",sqlo_rc);                /*@d14218tjv*/
      }

      /* Remove the semaphore set */

      sqlo_rc = sqloSSemSetRemove(&ipcConHandle->ipcResources.ipcSemSetID);

      if (sqlo_rc != SQLO_OK)
      {
         sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCIPCTERM, 14, 8,
                       (void *)&ipcConHandle->ipcResources.ipcSemSetID);

         rc = SQLCC_ERR_SYSTEM_ERROR;
         ipcseterr(rc,8,"SEMAPHORE REMOVE",sqlo_rc);              /*@d14218tjv*/
      }

   }
   else if (ipcConHandle->ipcResources.Allocated)                 /*@d38558tjv*/
   {
      /* We are executing on the client detach the shared memory set */

      sqlo_rc = sqlodshr(&ipcConHandle->ipcResources.ipcShmHandle);

      if (sqlo_rc != SQLO_OK)
      {
         sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCIPCTERM, 16,
                          sizeof(ipcConHandle->ipcResources.ipcShmHandle),
                          (void *)&ipcConHandle->ipcResources.ipcShmHandle);

         rc = SQLCC_ERR_SYSTEM_ERROR;
         ipcseterr(rc,8,"DETACH SHARED MEMORY",sqlo_rc);          /*@d14218tjv*/
      }
   }

   /* Free the IPC protocol specific connection handle */

   sqlo_rc = sqlofmblk(NULL, ipcConHandle);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCIPCTERM, 18,
                       sizeof(ipcConHandle),
                       (void *)&ipcConHandle);

      rc = SQLCC_ERR_SYSTEM_ERROR;
   }

exit:

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCTERM, (int) rc);
   return(rc);
}

/*****************************************************************************
*
*  Function Name     = sqlccipcdariterm
*
*  Component Name    = SQLCC
*
*  Descriptive Name  = IPC protocol support termination for an agent running
*                      a DARI
*
*  Func. Prototype   = short sqlccipctermDARI(
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             SQLCC_COND_T          *pCondition
*                                      )
*
*  Function          = This function performs resource clean-up task
*                      for a disconnected IPC connection.  It frees
*                      up all IPC specific data structures acquired
*                      during the sqlccipcinit processing, and also
*                      disconnects the process from the IPC resources.
*
*  Input             = 1) pComHandle: a pointer to the Communication
*                                     Handle that contains pointers
*                                     to all the IPC specific
*                                     functions and data structures.
*                                     All these data structures
*                                     are freed by this function.
*
*  Output            = 1) pCondition: a pointer to the common Condition
*                                     structure that contains a pointer
*                                     to the IPC specific condition
*                                     structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_MEM_NOT_FREED
*                      SQLCC_ERR_SYSTEM_ERROR
*                      SQLCC_ERR_NO_CONNECTION
*
*****************************************************************************/
                                                                    /*@fb12409SAM*/
short sqlccipcdariterm ( SQLCC_COMHANDLE_T   * pComHandle,
                         SQLCC_COND_T        * pCondition )

{
   short        rc      = SQLCC_OK;
   int          sqlo_rc = SQLO_OK;
   SQLCC_IPC_CON_HANDLE_T * ipcConHandle;

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCDARITERM);

   /* Ensure that we have a connection */

   if (pComHandle->connHandle.pConnHandleProto == NULL)
   {
      rc = SQLCC_ERR_NO_CONNECTION;
      goto exit;
   }

   /* Set up a local connection handle for readability */

   ipcConHandle = (SQLCC_IPC_CON_HANDLE_T *)
                  pComHandle->connHandle.pConnHandleProto;

   /* Close the receive semaphore, if the close fails set the */
   /* return code, issue a trace, and continue.               */

   sqlo_rc = sqloSSemClose(&ipcConHandle->ipcResources.ipcSemSetID,
                           &ipcConHandle->ipcResources.recvSemID);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCDARITERM, 1,
                     8, (void *)&ipcConHandle->ipcResources.ipcSemSetID,
                     8, (void *)&ipcConHandle->ipcResources.recvSemID);

      rc = SQLCC_ERR_SYSTEM_ERROR;
      ipcseterr(rc,8,"SEMAPHORE CLOSE",sqlo_rc);
   }

   /* Close the send semaphore, if the close fails set the */
   /* return code, issue a trace, and continue.            */

   sqlo_rc = sqloSSemClose(&ipcConHandle->ipcResources.ipcSemSetID,
                           &ipcConHandle->ipcResources.sendSemID);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCDARITERM, 2,
                     8, (void *)&ipcConHandle->ipcResources.ipcSemSetID,
                     8, (void *)&ipcConHandle->ipcResources.sendSemID);

      rc = SQLCC_ERR_SYSTEM_ERROR;
      ipcseterr(rc,8,"SEMAPHORE CLOSE",sqlo_rc);
   }

   /* Close the sync semaphore */                                   /*@bd15910SAM*/

   sqlo_rc = sqloSSemClose(&ipcConHandle->ipcResources.ipcSemSetID,
                           &ipcConHandle->ipcResources.syncSemID);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCDARITERM, 3,
                     8, (void *)&ipcConHandle->ipcResources.ipcSemSetID,
                     8, (void *)&ipcConHandle->ipcResources.syncSemID);

      rc = SQLCC_ERR_SYSTEM_ERROR;
      ipcseterr(rc,8,"SEMAPHORE CLOSE",sqlo_rc);
   }                                                                /*@ed15910SAM*/

   /* don't do a sqlodshr to detach the shared memory.  This is */  /*@d17343SAM*/
   /* unnecessary since FreeIPC will free it.  Otherwise we     */  /*@d17343SAM*/
   /* will try and free it and it will already be gone.         */  /*@d17343SAM*/


   /* Free the IPC protocol specific connection handle */

   sqlo_rc = sqlofmblk(NULL, ipcConHandle);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCIPCDARITERM, 6,
                       sizeof(ipcConHandle),
                       (void *)&ipcConHandle);

      rc = SQLCC_ERR_SYSTEM_ERROR;
   }

exit:

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCDARITERM, (int) rc);
   return(rc);
}
                                                                    /*@fe12409SAM*/

/*****************************************************************************
*
*  Function Name     = sqlccipcsend
*
*  Component Name    = SQLCC
*
*  Descriptive Name  = IPC protocol send data
*
*  Func. Prototype   = short sqlccipcsend(
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             SQLCC_COND_T          *pCondition
*                                      )
*
*  Function          = This function sends data on an established IPC
*                      connection.  If the sends a buffer and specifies
*                      the size of the buffer to be equal to the actual
*                      shared buffer size then we will set the packet
*                      switch flag on.  This flag will indicate to the
*                      receiver that the actual request spans more than
*                      one communication buffer.  The receiver will
*                      allocate a private buffer and copy the data into
*                      the private buffer before posting completion to the
*                      sender.
*
*                      NOTE: This mechanism assumes that the sender only
*                            uses n-1 bytes of the buffer where n is
*                            equivalent to the buffer size.  If the data
*                            to be sent exceeds n-1 bytes the bytes sent is
*                            set to n to indicate that there will be another
*                            buffer.
*
*  Input             = 1) pComHandle: a pointer to the Communication
*                                     Handle that contains pointers
*                                     to all the TCP/IP specific
*                                     functions and data structures.
*
*  Output            = 1) pCondition: a pointer to the common Condition
*                                     structure that contains a pointer
*                                     to the TCP/IP specific condition
*                                     structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_CONN_CLOSED_BY_PARTNER             @d14218tjv
*                      SQLCC_ERR_NO_CONNECTION
*                      SQLCC_ERR_INV_DATA_LENGTH
*
*****************************************************************************/

short sqlccipcsend( SQLCC_COMHANDLE_T     *pComHandle,
                    SQLCC_COND_T          *pCondition)

{
   short        rc      = SQLCC_OK;
   int          sqlo_rc = SQLO_OK;
   char         packetMode;                                       /*@d14857tjv*/

   SQLCC_IPC_CON_HANDLE_T * ipcConHandle;

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCSEND);

   /* Set up a local connection handle for readability */

   ipcConHandle = (SQLCC_IPC_CON_HANDLE_T *)
                  pComHandle->connHandle.pConnHandleProto;

   if ((ipcConHandle->serverFlag == FALSE) &&                    /*@bd15397mah*/
       (ipcConHandle->aslCommHeader->agentGoingDown == TRUE))
   {
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;
      ipcseterr(rc,8,"AGENT FORCED",sqlo_rc);
      goto exit;
   }                                                             /*@ed15397mah*/

   /* Ensure that the data length is less than or equal to the */
   /* shared memory buffer size.                               */
   if (pComHandle->comSend.ulDataLen > ipcConHandle->aslHeapSz)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCSEND, 1,
                     sizeof(pComHandle->comSend.ulDataLen),
                     (void *)&pComHandle->comSend.ulDataLen,
                     sizeof(ipcConHandle->aslHeapSz),
                     (void *)&ipcConHandle->aslHeapSz);

      rc = SQLCC_ERR_INV_DATA_LENGTH;
      goto exit;
   }

   sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCSEND, 2,
                  4, (void *)&pComHandle->comSend.pucBuffer,
                  4, (void *)&ipcConHandle->commBuffer);

   /* If the buffer passed in is a private memory buffer copy the */
   /* data into the shared communications buffer.                 */

   if (pComHandle->comSend.pucBuffer != ipcConHandle->commBuffer)
   {
      sqlomcpy(ipcConHandle->commBuffer,
               pComHandle->comSend.pucBuffer,
               pComHandle->comSend.ulDataLen);

   }

   if (pComHandle->comSend.usBriefTrace == SQLCC_TRUE)           /*@db23098tjv*/
   {
      sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCIPCSEND, 3,
                    SQLCC_BRIEF_TRACE_LEN,
                    (void *)pComHandle->comSend.pucBuffer);
   }
   else
   {
      sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCIPCSEND, 4,
                    pComHandle->comSend.ulDataLen,
                    (void *)pComHandle->comSend.pucBuffer);
   }                                                             /*@de23098tjv*/

   /* If the buffer size passed in is equal to the actual shared  */
   /* memory buffers size, and if we are not already in the       */
   /* packet switching mode, then turn on the packet switch flag. */

   if (pComHandle->comSend.ulDataLen == ipcConHandle->aslHeapSz) /*@d16619tjv*/
   {
      ipcConHandle->aslCommHeader->packetMode = TRUE;
   }
   else if (pComHandle->comSend.ulDataLen < ipcConHandle->aslHeapSz) /*@d16619tjv*/
   {
      /* We are currently in the packet switching mode, and the buffer */
      /* being sent is less than the size of the communication buffer. */
      /* Turn packet switching off.                                    */

      ipcConHandle->aslCommHeader->packetMode = FALSE;
   }

   /* Set the number of bytes being sent in the ASL buffer header */

   ipcConHandle->aslCommHeader->dataSent = pComHandle->comSend.ulDataLen;

   /* Turn off the receive termination flag */

   ipcConHandle->aslCommHeader->recvTermFlag = FALSE;


   /* If we are executing on the server there is a condition  */ /*@db13196tjv*/
   /* where the client may have exited abnormally while the   */
   /* agent is processing its request.  If this condition     */
   /* occurs the next receive on the server will succeeed due */
   /* to the UNDO operation.  The server receive thus needs a */
   /* flag to indicate whether the send originated at the     */
   /* client or the server.                                   */

   if (ipcConHandle->serverFlag == TRUE)
   {
      ipcConHandle->aslCommHeader->serverSend = TRUE;
   }
   else
   {
      // We are on the client, the server may go down              @bd66222tjv
      // before it receives the request.  This will
      // remove the IPC resources and the recvTermFlag
      // will be FALSE as set above.  This will cause
      // the client receive logic to think that the
      // IPC resources were intentionally removed by
      // the server and will the receive the SQLCA
      // as sent by the server.  We need to be able to
      // detect the condition that the IPC resource were
      // removed as a result of a trap in the instance.
      // We will do this by assuming a trap will occur
      // and set the agentGoingDown flag to TRUE.  It
      // will get set to FALSE if the server receives
      // the request and resets the recvTermFlag to
      // TRUE.
      ipcConHandle->aslCommHeader->agentGoingDown = TRUE;        //@ed66222tjv
      ipcConHandle->aslCommHeader->serverSend = FALSE;
   }                                                             /*@de13196tjv*/

   /* A timing situation exists where this process may lose */   /*@db14857tjv*/
   /* its time slice immediately after posting the send     */
   /* semaphore.  The receiving process can then get the    */
   /* next time slice, receive the buffer and then perform  */
   /* a send.  If this send sets on the global packetMode   */
   /* flag before we get back our time slice we will fall   */
   /* into the wait on the packet switch semaphore test.    */
   /* This would then decrement the semaphore post which    */
   /* was actually intended for the receiver.  Save a local */
   /* copy of the semaphore before performing the post.     */

   packetMode = ipcConHandle->aslCommHeader->packetMode;         /*@de14857tjv*/

   /* Post the send semaphore.  Note that the send semaphore          */
   /* actually differs on the client and the server.  On the server   */
   /* the receive semaphore actually corresponds to the client's      */
   /* send semaphore, and vice versa.  These semaphores are reversed  */
   /* to ensure that the client only ever waits on one semaphore, and */
   /* the server the other.  This ensures that we will never have a   */
   /* situation where the client can perform a send, and followed     */
   /* immediately by a receive which wakes up on the semaphore posted */
   /* by the send.                                                    */

   sqlo_rc = sqloSSemV (&ipcConHandle->ipcResources.ipcSemSetID,
                        &ipcConHandle->ipcResources.sendSemID, 0);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_fnc_data2 (SQLT_SQLCC, SQLT_SQLCCIPCSEND, 5,
                      sizeof(ipcConHandle->ipcResources.ipcSemSetID),
                      (void *)&ipcConHandle->ipcResources.ipcSemSetID,
                      sizeof(ipcConHandle->ipcResources.sendSemID),
                      (void *)&ipcConHandle->ipcResources.sendSemID);
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                      /*@d14218tjv*/
      ipcseterr(rc,8,"SEMAPHORE POST",sqlo_rc);                   /*@d14218tjv*/
      goto exit;
   }

   /* Wait on the receive semaphore, if the packet switch flag is on. */
   /* This ensures that the sending process does not overwrite the    */
   /* shared communication buffer before the receiver copies the      */
   /* data into private storage.                                      */

   if (packetMode == TRUE)                                        /*@d14857tjv*/
   {
      sqlo_rc = sqloSSemP (&ipcConHandle->ipcResources.ipcSemSetID,
                           &ipcConHandle->ipcResources.recvSemID,
                           0,
                           (U32BIT)SQLO_NO_TIMEOUT);              /*@d120677pj*/

      if (sqlo_rc != SQLO_OK)
      {
         sqlt_fnc_data2 (SQLT_SQLCC, SQLT_SQLCCIPCSEND, 6,
                         sizeof(ipcConHandle->ipcResources.ipcSemSetID),
                         (void *)&ipcConHandle->ipcResources.ipcSemSetID,
                         sizeof(ipcConHandle->ipcResources.recvSemID),
                         (void *)&ipcConHandle->ipcResources.recvSemID);
         rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                   /*@d14218tjv*/
         ipcseterr(rc,8,"SEMAPHORE WAIT",sqlo_rc);                /*@d14218tjv*/
         goto exit;
      }
   }

exit:

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCSEND, (int) rc);
   return(rc);
}

/*****************************************************************************
*
*  Function Name     = sqlccipcrecv
*
*  Component Name    = SQLC
*
*  Descriptive Name  = IPC protocol receive data
*
*  Func. Prototype   = short sqlccipcrecv(
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             SQLCC_COND_T          *pCondition
*                                      )
*
*  Function          = This function receives data from an established
*                      IPC connection.  If the packet switch flag is set
*                      on the receive function will allocate a private
*                      communication buffer (if one has not been allocated).
*                      If a private buffer was allocate and the packet
*                      switch goes off then deallocate the private buffer.
*
*                      In some cases the user may use a private buffer,
*                      when this occurs there is no need to allocate or
*                      deallocate a private buffer for packet switching.
*
*  Input             = 1) pComHandle: a pointer to the Communication
*                                     Handle that contains pointers
*                                     to all the IPC specific
*                                     functions and data structures.
*
*  Output            = 1) pCondition: a pointer to the common Condition
*                                     structure that contains a pointer
*                                     to the IPC specific condition
*                                     structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_CONN_CLOSED_BY_PARTNER
*                      SQLCC_ERR_NO_CONNECTION
*                      SQLCC_ERR_REQUEST_INTERRUPTED
*                      SQLCC_ERR_RECV_BUFFER_TOO_SMALL
*                      SQLCC_ERR_SYSTEM_ERROR
*                      SQLCC_ERR_MEM_NOT_FREED
*
*****************************************************************************/

short sqlccipcrecv ( SQLCC_COMHANDLE_T      *pComHandle,
                     SQLCC_COND_T           *pCondition )

{
   short        rc      = SQLCC_OK;
   int          sqlo_rc = SQLO_OK;

   SQLCC_IPC_CON_HANDLE_T * ipcConHandle;
   U32BIT                   semFlag = 0;

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCRECV);

   /* Ensure that we have a connection */

   /* Set up a local connection handle for readability */

   ipcConHandle = (SQLCC_IPC_CON_HANDLE_T *)
                  pComHandle->connHandle.pConnHandleProto;

   sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCRECV, 1,
                  4, (void *)pComHandle->comRecv.pucBuffer,
                  4, (void *)&ipcConHandle->commBuffer);

   /* Wait on the receive semaphore.  As with the send semaphore the    */
   /* receive semaphore on the client corresponds to the send semaphore */
   /* on ther server and vice versa.  This is to ensure that only the   */
   /* opposite process will wake up on the semaphore.                   */

   if (pComHandle->comRecv.usIgnoreSigint == SQLCC_FALSE)
   {
      semFlag = SQLO_SSEM_INTERRUPTIBLE;
      sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCIPCRECV, 55,
                    sizeof(semFlag), (void *)&semFlag);
   }

   if (pComHandle->comRecv.asyncmode == SQLCC_FALSE)             /*@db40097cia*/
   {
      #if defined(NOKILLHANDLER)                                 //@ed80241tjv
      if (ipcConHandle->serverFlag == TRUE)
      {
         // We are on the server, on this system there is no way
         // to handle the case where an process has been killed.
         // Wake up every 2 minutes and determine if the application
         // still exists.
         for (;;) {
            sqlo_rc = sqloSSemP(&ipcConHandle->ipcResources.ipcSemSetID,
                                &ipcConHandle->ipcResources.recvSemID,
                                semFlag,
                                                                  //@db125200jon
                                ((pComHandle->comTimeout == 0) ? 120
                                                               : pComHandle->comTimeout));
                                                                  //@de125200jon

            if (sqlo_rc == SQLO_SSEM_TIMEOUT)
            {
               if (pComHandle->comTimeout > 0 &&                    //@db125200jon
                   ipcConHandle->isDariAgent != TRUE)              //@d138623swm
               {
                  rc = SQLCC_ERR_RECV_TIMEOUT;
                  goto exit;
               }                                                    //@de125200jon

               sqlo_rc = SQLOQPRC(ipcConHandle->aslCommHeader->clientPid);

               if (sqlo_rc == SQLO_PROC_NOT_FOUND)
               {
                  break;
               }
            }
            else
            {
               break;
            }
         }
         // If only select is to be done then return with status    //@db127829jon
         // that the socket is ready to be read.
         if ((pComHandle->comTimeout > 0)
             && (pComHandle->comSelectOnly == TRUE))
         {
            sqloSSemV(&ipcConHandle->ipcResources.ipcSemSetID,
                      &ipcConHandle->ipcResources.recvSemID,
                      semFlag);
            goto exit;
         }                                                          //@de127829jon
      }
      else
      {
         sqlo_rc = sqloSSemP(&ipcConHandle->ipcResources.ipcSemSetID,
                             &ipcConHandle->ipcResources.recvSemID,
                             semFlag,
                                                                    //@db125200jon
                             ((pComHandle->comTimeout == 0) ? SQLO_NO_TIMEOUT
                                                            : pComHandle->comTimeout));
         if (sqlo_rc == SQLO_SSEM_TIMEOUT)
         {
            rc = SQLCC_ERR_RECV_TIMEOUT;
            goto exit;
         }                                                          //@de125200jon
         // If only select is to be done then return with status    //@db127829jon
         // that the socket is ready to be read.
         if ((pComHandle->comTimeout > 0)
             && (pComHandle->comSelectOnly == TRUE))
         {
            sqloSSemV(&ipcConHandle->ipcResources.ipcSemSetID,
                      &ipcConHandle->ipcResources.recvSemID,
                      semFlag);
            goto exit;
         }                                                          //@de127829jon
      }
      #else
      sqlo_rc = sqloSSemP(&ipcConHandle->ipcResources.ipcSemSetID,
                          &ipcConHandle->ipcResources.recvSemID,
                          semFlag,
                          (U32BIT)((pComHandle->comTimeout == 0)    //@db125200jon
                                   ? SQLO_NO_TIMEOUT                /*@d120677pj*/
                                   : pComHandle->comTimeout));
      if (sqlo_rc == SQLO_SSEM_TIMEOUT)
      {
         rc = SQLCC_ERR_RECV_TIMEOUT;
         goto exit;
      }                                                             //@de125200jon
      // If only select is to be done then return with status       //@db127829jon
      // that the socket is ready to be read.
      if ((pComHandle->comTimeout > 0)
          && (pComHandle->comSelectOnly == TRUE))
      {
         sqloSSemV(&ipcConHandle->ipcResources.ipcSemSetID,
                   &ipcConHandle->ipcResources.recvSemID,
                   semFlag);
         goto exit;
      }                                                             //@de127829jon
      #endif                                                        //@ed80241tjv
   }
   else
   {
      sqlo_rc = sqloSSemP(&ipcConHandle->ipcResources.ipcSemSetID,
                          &ipcConHandle->ipcResources.recvSemID,
                          semFlag,
                          0);
      if (sqlo_rc == SQLO_SSEM_TIMEOUT)
      {
         rc = SQLE_RC_STILL_EXECUTING;
         goto exit;
      }
   }                                                             /*@de40097cia*/

   // If this is the server reset the agentGoingDown flag          @bd66222tjv
   // to FALSE to negate the setting of TRUE done by the
   // client send routine.  See the comment in sqlccipcsend
   // as to why we set the flag to TRUE in the first place.

   if (ipcConHandle->serverFlag == TRUE)
   {
      ipcConHandle->aslCommHeader->agentGoingDown = FALSE;
   }                                                             //@ed66222tjv

   /* There is a condition on the client where the server has issued a */
   /* send of an error condition followed by a disc and a term.  The   */
   /* term will remove the semaphore set thus causing a failure on     */
   /* the SSemP call.  When this situation occurs the recvTermFlag     */
   /* will be FALSE indicating that a send occurred from the server.   */
   /* Process this message, and return a normal condition.  The actual */
   /* error will either be processed from the error information in the */
   /* buffer sent from the server, or from the next send/recv.         */

   if (!((ipcConHandle->serverFlag == FALSE) &&
         (ipcConHandle->aslCommHeader->recvTermFlag == FALSE)) &&
         (sqlo_rc != SQLO_OK))
   {
      if ((pComHandle->comRecv.usIgnoreSigint == SQLCC_FALSE) &&
          (sqlo_rc == SQLO_USER_TERM))                            /*@d12153tjv*/
      {
         rc = SQLCC_ERR_REQUEST_INTERRUPTED;
      }
      else
      {
         sqlt_fnc_data2 (SQLT_SQLCC, SQLT_SQLCCIPCRECV, 2,
                         sizeof(ipcConHandle->ipcResources.ipcSemSetID),
                         (void *)&ipcConHandle->ipcResources.ipcSemSetID,
                         sizeof(ipcConHandle->ipcResources.recvSemID),
                         (void *)&ipcConHandle->ipcResources.recvSemID);
         rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                   /*@d14218tjv*/
         ipcseterr(rc,8,"SEMAPHORE WAIT",sqlo_rc);                /*@d14218tjv*/
      }
      goto exit;
   }

   /* If the receive termination flag is set our partner died horribly */

   if (ipcConHandle->aslCommHeader->recvTermFlag == TRUE)
   {
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                      /*@d14218tjv*/
      goto exit;
   }

   /* If the disconnect flag is set on our partner performed a disconnect */

   if (ipcConHandle->aslCommHeader->disconnectRqst == TRUE)
   {
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;
      goto exit;
   }

   /* If we are executing on the server and the serverSend   */  /*@db13196tjv*/
   /* flag is set we are receiving the send from the server, */
   /* this indicates that the client has exited abnormally.  */

   if ((ipcConHandle->serverFlag == TRUE) &&
       (ipcConHandle->aslCommHeader->serverSend == TRUE))
   {
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                      /*@d14218tjv*/
      ipcseterr(rc,8,"serverSend",sqlo_rc);
      goto exit;
   }                                                             /*@de13196tjv*/
   /* If we're the client, there is not a fresh send from the      @bd13487mah*/
   /* server (serverSend==FALSE), and the agent is going down,                */
   /* report an error.                                                        */
   else if ((ipcConHandle->serverFlag == FALSE) &&
            (ipcConHandle->aslCommHeader->serverSend == FALSE) &&
            (ipcConHandle->aslCommHeader->agentGoingDown == TRUE))
   {
      rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                     /*d14425mah*/
      ipcseterr(rc,8,"AGENT FORCED",sqlo_rc);                    /*@dxxxxtjv*/
      goto exit;
   }                                                             /*@bd13487mah*/

   if (pComHandle->comRecv.usBriefTrace == SQLCC_TRUE)           /*@db23098tjv*/
   {
      sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCIPCRECV, 3,
                    SQLCC_BRIEF_TRACE_LEN,
                    (void *)ipcConHandle->commBuffer);
   }
   else
   {
      sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCIPCRECV, 4,
                    ipcConHandle->aslCommHeader->dataSent,
                    (void *)ipcConHandle->commBuffer);
   }                                                             /*@de23098tjv*/

   /* Set the bytes received field in the communication handle */

   pComHandle->comRecv.ulBytesRcvd = ipcConHandle->aslCommHeader->dataSent;

   /* If the user provided us with a private communications buffer */
   /* copy the data from the shared buffer into the private buffer */

   if (ipcConHandle->commBuffer != *pComHandle->comRecv.pucBuffer)
   {
      /* Ensure that the user's buffer is equivalent to the */
      /* shared communication buffer size.                  */

      if (pComHandle->comRecv.ulDataLen != ipcConHandle->aslHeapSz)
      {
         sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCRECV, 5,
                        sizeof(pComHandle->comRecv.ulDataLen),
                        (void *)&pComHandle->comRecv.ulDataLen,
                        sizeof(ipcConHandle->aslHeapSz),
                        (void *)&ipcConHandle->aslHeapSz);

         rc = SQLCC_ERR_RECV_BUFFER_TOO_SMALL;
         pComHandle->comRecv.ulBytesRcvd = 0;
         goto exit;
      }

      /* The buffer may be the private buffer which was allocated to */
      /* store the data in the shared communication buffer when we   */
      /* switch into the packet switch mode.  If the packet switch   */
      /* flag has been turned off then we should deallocate the      */
      /* private buffer and simply return the shared communication   */
      /* buffer address.                                             */
      if ((*pComHandle->comRecv.pucBuffer == ipcConHandle->privateBuffer) &&
          (ipcConHandle->aslCommHeader->packetMode == FALSE))

      {
         sqlo_rc = sqlofmblk(NULL, ipcConHandle->privateBuffer);
         if (sqlo_rc != SQLO_OK)
         {
            sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCIPCRECV, 6,
                             sizeof(ipcConHandle->privateBuffer),
                             (void *)&ipcConHandle->privateBuffer);

            rc = SQLCC_ERR_MEM_NOT_FREED;
            pComHandle->comRecv.ulBytesRcvd = 0;
            goto exit;
         }
         ipcConHandle->privateBuffer = NULL;
         *pComHandle->comRecv.pucBuffer = ipcConHandle->commBuffer; /*@d14036tjv*/
      }
      else
      {
         sqlomcpy(*pComHandle->comRecv.pucBuffer,
                  ipcConHandle->commBuffer,
                  ipcConHandle->aslCommHeader->dataSent);
      }
   }

   /* The user passed in a pointer to the shared memory communication */
   /* buffer.  If the packet switch mode is on allocate a private     */
   /* buffer, copy the data into that buffer, and return the pointer  */
   /* to the private communication buffer.                            */

   else if (ipcConHandle->aslCommHeader->packetMode == TRUE)
   {
      sqlo_rc = sqlogblk(NULL,
                         ipcConHandle->aslHeapSz,
                         (void **)&ipcConHandle->privateBuffer);

      if (sqlo_rc != SQLO_OK)
      {
         if (sqlo_rc == SQLO_NOMEM)
         {
            rc = SQLCC_ERR_MEM_UNAVAIL;
         }
         else
         {
            rc = SQLCC_ERR_SYSTEM_ERROR;
         }
         goto exit;
      }

      *pComHandle->comRecv.pucBuffer = ipcConHandle->privateBuffer;

      sqlomcpy(*pComHandle->comRecv.pucBuffer,
               ipcConHandle->commBuffer,
               ipcConHandle->aslCommHeader->dataSent);

   }

   /* Reset the receive termination flag */

   ipcConHandle->aslCommHeader->recvTermFlag = TRUE;

   if (ipcConHandle->aslCommHeader->packetMode == TRUE)
   {
      /* Post the send semaphore to acknowledge that the receive   */
      /* has completed and copied the data into the private buffer */

      sqlo_rc = sqloSSemV(&ipcConHandle->ipcResources.ipcSemSetID,
                          &ipcConHandle->ipcResources.sendSemID, 0);

      if (sqlo_rc != SQLO_OK)
      {
         sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCIPCRECV, 7,
                        sizeof(ipcConHandle->ipcResources.ipcSemSetID),
                        (void *)&ipcConHandle->ipcResources.ipcSemSetID,
                        sizeof(ipcConHandle->ipcResources.sendSemID),
                        (void *)&ipcConHandle->ipcResources.sendSemID);

         rc = SQLCC_ERR_CONN_CLOSED_BY_PARTNER;                   /*@d14218tjv*/
         ipcseterr(rc,8,"SEMAPHORE POST",sqlo_rc);                /*@d14218tjv*/
         goto exit;
      }
   }

exit:

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCRECV, (int) rc);
   return(rc);
}

/*****************************************************************************
*
*  Function Name     = sqlccipcintrr
*
*  Component Name    = SQLC
*
*  Descriptive Name  = IPC protocol interrupt current SQL request
*
*  Func. Prototype   = short sqlccipcintrr(
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             SQLCC_COND_T          *pCondition
*                                       )
*
*  Function          = This function forwards an interrupt request to the
*                      IPC listener associated with the current connection
*
*  Input             = 1) pComHandle: a pointer to the Communication
*                                     Handle that contains pointers
*                                     to all the IPC specific
*                                     functions and data structures.
*
*  Output            = 1) pCondition: a pointer to the common Condition
*                                     structure that contains a pointer
*                                     to the IPC specific condition
*                                     structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_COMM_REQUEST_FAILED
*                      SQLCC_ERR_NO_CONNECTION
*
*****************************************************************************/

short sqlccipcintrr( SQLCC_COMHANDLE_T     *pComHandle,
                     SQLCC_COND_T          *pCondition )

{
   short                rc      = SQLCC_OK;
   int                  sqlo_rc = SQLO_OK;
   SQLE_DB2RA_T         *t_db2ra;                                 /*@d14521tjv*/
   unsigned short       *agent_index;                             /*@d14521tjv*/

   SQLCC_IPC_INTR_REQST_T   intr_request;
   SQLCC_IPC_INTR_REPLY_T   intr_reply;

   SQLER_GLOB * erglob_ptr;                                       //@bd80150tjv
   struct sql_static_data * pGp = sqlo_get_static_data();
   erglob_ptr = &pGp->pstSqleUbi->s_sqlerglob;                    //@ed80150tjv

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCINTR);

   /* Set a pointer to the DB2RA in the communication buffer. */ /*@db14521tjv*/
   /* This pointer proceed the 4 byte length, the 4 byte data */
   /* sent, and the 4 byte alloc size.                        */

   t_db2ra = (SQLE_DB2RA_T *)(pComHandle->comInterrupt.pucBuffer + 12);

   /* Copy the agent index out of the buffer, it follows */
   /* the DB2RA in the buffer.                           */

   agent_index = (unsigned short *)
                 (pComHandle->comInterrupt.pucBuffer + 12 +
                 t_db2ra->sqlrabc);
                                                                 /*@db14521tjv*/
   intr_request.msg_hdr.msg_buf_size =
                            sizeof(intr_request) -
                            sizeof(intr_request.msg_hdr.msg_buf_size);
                                                                 /*@de14521tjv*/

   intr_request.type                 = SQLCC_IPC_INTERRUPT;

   /* Copy the agent index out of the buffer.  This index is  */  /*@d14521tjv*/
   /* preceeded by the 4 byte CCI length, and the DB2RA block */  /*@d14521tjv*/
   sqlomcpy(&intr_request.agent_index,                            /*@d14521tjv*/
            agent_index, sizeof(*agent_index));                   /*@d23502mah*/

                                                                 /*@db14521tjv*/
   intr_reply.msg_hdr.msg_buf_size = sizeof(intr_reply) -
                            sizeof(intr_reply.msg_hdr.msg_buf_size);
                                                                 /*@de14521tjv*/

   /* Set the interrupt flag in the global communications buffer */

   ((SQLCC_IPC_CON_HANDLE_T *)
    pComHandle->connHandle.pConnHandleProto)->aslCommHeader->interruptFlag =
                                                                         TRUE;

   /* NOTE: need to add the instance name to this API.  If the */
   /*       NECB has no instance name leave as NULL.           */

   sqlo_rc = sqloTQMRequestMLN(                                  /*@d44244jkl*/
                    ((SQLCC_IPC_CON_HANDLE_T *)
                    pComHandle->connHandle.pConnHandleProto)->instanceName,
                    SQLO_QUE_IPCLIST,
                    (SQLO_MSG_HDR *)&intr_request,
                    SQLOQ_HIGH,
                    (SQLO_MSG_HDR *)&intr_reply,
                    SQLO_QUE_WAIT | SQLO_QUE_NO_TRUN,
                    SQLO_NO_TIMEOUT,
                    sqlz_krcbp->sqlekcon.nodetype,               /*@d44244jkl*/
                    erglob_ptr->lConnAttNode);                   //@d80150xtjv

   if (sqlo_rc != SQLO_OK)
   {
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
      ipcseterr(rc,8,"QUEUE WRITE",sqlo_rc);                      /*@d14218tjv*/
   }

exit:

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCINTR, (int) rc);
   return(rc);
}

/*****************************************************************************
*
*  Function Name     = sqlccipcgetapplid
*
*  Component Name    = SQLCC
*
*  Descriptive Name  = IPC protocol get application ID
*
*  Func. Prototype   = short sqlccipcgetapplid(
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             char                  *pucApplId,
*                             SQLCC_COND_T          *pCondition
*                                       )
*
*  Function          = This function generates an application ID for local
*                      IPC based connections.  The format of an application
*                      ID for an IPC connection is:
*
*                               *LOCAL.instance_name.timestamp
*
*                      Where:
*                               instance_name   => value set in DB2INSTANCE
*                               timestamp       => generated by OSS
*
*  Input             = 1) pComHandle: a pointer to the Communication
*                                     Handle that contains pointers
*                                     to all the IPC specific
*                                     functions and data structures.
*
*  Output            = 1) pucApplId : Generate application ID
*                      2) pCondition: a pointer to the common Condition
*                                     structure that contains a pointer
*                                     to the IPC specific condition
*                                     structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_SYSTEM_ERROR
*                      SQLCC_ERR_NO_CONNECTION
*
*****************************************************************************/

short sqlccipcgetapplid ( SQLCC_COMHANDLE_T     *pComHandle,
                          char                  *pucApplId,
                          SQLCC_COND_T          *pCondition )

{
   short        rc      = SQLCC_OK;
   int          sqlo_rc = SQLO_OK;
   int          instanceNameLen;
   sqlint32     mynode = SQLZ_PDB_MY_NODE_NUM(sqlz_krcbp);         //@d190698swm
   char         nodenum[4];                                        //@d190698swm
   char         timestamp[13];

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCGETAPPLID);

   /* Copy the LOCAL prefix into the application identifier */
   if (SQLPDB_RT)
   {
     // Set to Nxxx where xxx is the node number                    @db190698swm
     sqlomcpy(pucApplId, "*N", sizeof("*N")-1);
     pucApplId += sizeof("*N")-1;
     sprintf(nodenum, "%d", mynode);
     sqlomcpy(pucApplId, nodenum, strlen(nodenum));
     pucApplId += strlen(nodenum);

     // Append a "."
     sqlomcpy(pucApplId, ".", sizeof(".")-1);
     pucApplId++;
   } else {
     sqlomcpy(pucApplId, "*LOCAL.", sizeof("*LOCAL.")-1);
     pucApplId += sizeof("*LOCAL.")-1;
   }                                                              //@de190698swm

   /* Copy the instance name into the application identifier */

   instanceNameLen =
            strlen(((SQLCC_IPC_CON_HANDLE_T *)
              pComHandle->connHandle.pConnHandleProto)->instanceName);

   sqlomcpy(pucApplId,
            ((SQLCC_IPC_CON_HANDLE_T *)
              pComHandle->connHandle.pConnHandleProto)->instanceName,
            instanceNameLen);

   pucApplId += instanceNameLen;
   sqlomcpy(pucApplId, ".", sizeof(".")-1);
   pucApplId++;

   /* Generate a timestamp and copy it into the application ID */

   sqlo_rc = sqlotimeappid(timestamp);

   if (sqlo_rc != SQLO_OK)
   {
      rc = SQLCC_ERR_SYSTEM_ERROR;
      goto exit;
   }

   sqlomcpy(pucApplId, timestamp, sizeof(timestamp));

   *(pucApplId+sizeof(timestamp)) = '\0';

exit:

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCGETAPPLID, (int) rc);
   return(rc);
}

/*****************************************************************************
*
*  Function Name     = sqlccipcputapplid
*
*  Component Name    = SQLCC
*
*  Descriptive Name  = IPC protocol put application ID
*
*  Func. Prototype   = short sqlccipcputapplid(
*                             SQLCC_COMHANDLE_T     *pComHandle,
*                             char                  *pucApplId,
*                             SQLCC_COND_T          *pCondition
*                                       )
*
*  Function          = This function takes an application ID and stores the
*                      location information in the connection handle.  The
*                      application ID should be of the form:
*
*                               *LOCAL.instance_name.timestamp
*
*                      The location is one of the tkens returned when a
*                      communications error occurs, and a 30081 is returned.
*                      This function should only be called from the client.
*
*  Input             = 1) pComHandle: a pointer to the Communication
*                                     Handle that contains pointers
*                                     to all the IPC specific
*                                     functions and data structures.
*                      2) pucApplId : application ID
*
*  Output            = 1) pCondition: a pointer to the common condition
*                                     structure that contains a pointer
*                                     to the IPC specific condition
*                                      structure.
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_SYSTEM_ERROR
*                      SQLCC_ERR_NO_CONNECTION
*
*****************************************************************************/

short sqlccipcputapplid( SQLCC_COMHANDLE_T    *pComHandle,
                         char                 *pucApplId,
                         SQLCC_COND_T         *pCondition )

{
   short            rc          = SQLCC_OK;
   char            *ucFirstDot  = NULL;
   char            *ucSecondDot = NULL;

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCIPCPUTAPPLID );

   ucFirstDot = strchr(pucApplId, '.');

   if (ucFirstDot == NULL)
   {
      rc = SQLCC_ERR_INV_APPLID;
      goto exit;
   }

   ucSecondDot = strchr(ucFirstDot + 1, '.');

   if (ucSecondDot == NULL)
   {
      rc = SQLCC_ERR_INV_APPLID;
      goto exit;
   }

   sqlomcpy(((SQLCC_IPC_CON_HANDLE_T *)
             pComHandle->connHandle.pConnHandleProto)->commHostId,
             ucFirstDot + 1,
             ucSecondDot - ucFirstDot);

exit:

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCIPCPUTAPPLID, (int) rc );
   return (rc);
}
                                                                 /*@fe11325tjv*/

                                                                 /*@f11954SAM*/
/*****************************************************************************/
/* Function name: sqlccIPCtoASCII                                            */
/* Function:                                                                 */
/*         This function converts the IPC info structure into an ASIII       */
/*         string to be passed to the DARI when it is created.               */
/* Input:                                                                    */
/*         psCommInfo - IPC comm information                                 */
/* Output:                                                                   */
/*         pCmdLineArgs -  ASCII string                                      */
/*****************************************************************************/

short  sqlccIPCtoASCII ( SQLCC_INITSTRUCT_T   * psCommInfo,
                         char                 * pCmdLineArgs)
{
  int   rc = SQL_RC_OK;
  char  buffer[9];                                                  /*d14558SAM*/
  char  sqoCmdLineArgs[SQLE_DARI_CMDLINE_SZ];
  SQLCC_IPC_RESOURCES_T * pResources;

  sqlt_cei_entry(SQLT_SQLCC, SQLT_SQLCC_IPC_TO_ASCII);

  sqlomset(buffer, '\0', 9);                                        /*d14558SAM*/
  sqlomset(sqoCmdLineArgs, '\0', sizeof(sqoCmdLineArgs) );          /*d14558SAM*/

  pResources = &(psCommInfo->ipcResources);

  sprintf (buffer,"%x",pResources->ipcShmSize);
  strcat(pCmdLineArgs, buffer);
  strcat(pCmdLineArgs, ",");
 #ifndef SQLWINT  /* d21159djh */
  sprintf (buffer,"%x",pResources->syncSemID);
  strcat(pCmdLineArgs, buffer);
  strcat(pCmdLineArgs, ",");

  sprintf (buffer,"%x",pResources->recvSemID);
  strcat(pCmdLineArgs, buffer);
  strcat(pCmdLineArgs, ",");

  sprintf (buffer,"%x",pResources->sendSemID);
  strcat(pCmdLineArgs, buffer);
  strcat(pCmdLineArgs, ",");

  /* convert the sqo structures to an ASCII string */
  sqloIPCtoASCII ( &(pResources->ipcShmHandle), &(pResources->ipcSemSetID), sqoCmdLineArgs );
  strcat(pCmdLineArgs, sqoCmdLineArgs);
  #else        /*d21159djh*/
  sqloIPCtoASCII ( &(pResources->syncSemID), SQLOTYPE_SSEM,sqoCmdLineArgs );
  strcat(pCmdLineArgs, sqoCmdLineArgs);
  sqloIPCtoASCII ( &(pResources->recvSemID), SQLOTYPE_SSEM,sqoCmdLineArgs );
  strcat(pCmdLineArgs, sqoCmdLineArgs);
  sqloIPCtoASCII ( &(pResources->sendSemID), SQLOTYPE_SSEM,sqoCmdLineArgs );
  strcat(pCmdLineArgs, sqoCmdLineArgs);
  sqloIPCtoASCII ( &(pResources->ipcShmHandle), SQLOTYPE_MEM_SET,sqoCmdLineArgs );
  strcat(pCmdLineArgs, sqoCmdLineArgs);
  sqloIPCtoASCII ( &(pResources->ipcSemSetID), SQLOTYPE_SSEMSETID ,sqoCmdLineArgs );
  strcat(pCmdLineArgs, sqoCmdLineArgs);

  #endif

  sqlt_cei_retcode(SQLT_SQLCC, SQLT_SQLCC_IPC_TO_ASCII, rc);
  return (rc);
}

                                                                 /*@f11954SAM*/
/*****************************************************************************/
/* Function name: sqlccASCIItoIPC                                            */
/* Function:                                                                 */
/*         This function takes an ASCII string, and converts it into an      */
/*         IPC info structure.  Used to convert the DARI create command line.*/
/* Input:                                                                    */
/*         pCmdLineArgs -  ASCII string                                      */
/* Output:                                                                   */
/*         psCommInfo - IPC comm information                                 */
/*****************************************************************************/
 short  sqlccASCIItoIPC ( SQLCC_INITSTRUCT_T   * psCommInfo,
                          char                 * pCmdLineArgs)
{
  int  rc = SQL_RC_OK;
  char *end;              /* position after token */
  SQLCC_IPC_RESOURCES_T * pResources;

  sqlt_cei_entry(SQLT_SQLCC, SQLT_SQLCC_ASCII_TO_IPC);

  pResources = &(psCommInfo->ipcResources);

  /* decipher command line arguments */

  pResources->ipcShmSize                   = (U32BIT       ) strtoul (pCmdLineArgs, &end, 16);
#ifndef SQLWINT    /*d21159djh*/
  pResources->syncSemID                    = (SQLO_SSEMID) strtoul (++end, &end, 16);
  pResources->recvSemID                    = (SQLO_SSEMID) strtoul (++end, &end, 16);
  pResources->sendSemID                    = (SQLO_SSEMID) strtoul (++end, &end, 16);

  /* get the values for the SQO structures */
  sqloASCIItoIPC ( &(pResources->ipcShmHandle), &(pResources->ipcSemSetID), ++end );
#else           /*d21159djh*/
  end = sqloASCIItoIPC ( &(pResources->syncSemID),SQLOTYPE_SSEM, ++end );
  end = sqloASCIItoIPC ( &(pResources->recvSemID),SQLOTYPE_SSEM, ++end );
  end = sqloASCIItoIPC ( &(pResources->sendSemID),SQLOTYPE_SSEM, ++end );
  end = sqloASCIItoIPC ( &(pResources->ipcShmHandle),SQLOTYPE_MEM_SET, ++end );
  end = sqloASCIItoIPC ( &(pResources->ipcSemSetID),SQLOTYPE_SSEMSETID, ++end );
#endif

  pResources->sendRecvSemOpen = FALSE;                            /*@d45810tjv*/

  sqlt_cei_retcode(SQLT_SQLCC, SQLT_SQLCC_ASCII_TO_IPC, rc);
  return (rc);
}

                                                                 /*@bd13843mah*/
/**************************************************************************
*
*  Function Name  = sqlccipcAgentExitList
*  Descriptive Name = IPC specific Agent Exit List Handler
*  Function = Tell the client that the agent is dieing abnormally by setting
*             the recvTermFlag and posting the server's send semaphore.
*             This routine should only be called in the event of an abnormal
*             termination.
*  Dependencies = pConnHandleProto in the EDU static data area must be set up
*  Restrictions = None.
*  Input =  1. iTermCode - agent termination code
*  Output = None
*  Normal Return = None
*  Error Return = None
*
**************************************************************************/
void sqlccipcAgentExitList ( int iTermCode )
{

  int sqlo_rc;

  sqlt_cei_entry(SQLT_SQLC, SQLT_SQLCC_IPC_AGENT_EXIT_LIST);

  SQLCC_IPC_CON_HANDLE_T * ipcConHandle;

  struct sql_static_data * pGp = sqlo_get_static_data();

  ipcConHandle = (SQLCC_IPC_CON_HANDLE_T *) pGp->pstSqlc->pConnHandleProto;

  /* Set the AgentgoingDownFlag.  The client looks at this before recieving   */
  /* data.  This ensures that the client will know if the agent is going down */
  /* prematurely as a result of a FORCE or a trap in the engine.              */

  ipcConHandle->aslCommHeader->agentGoingDown = TRUE;

  /* Post semaphore that the client is waiting on */

  sqlo_rc = sqloSSemV (&ipcConHandle->ipcResources.ipcSemSetID,
                       &ipcConHandle->ipcResources.sendSemID, 0);
  if (sqlo_rc != SQLO_OK)
  {
     sqlt_system_error (SQLT_SQLCC, SQLT_SQLCC_IPC_AGENT_EXIT_LIST, 2,
                     sizeof(ipcConHandle->ipcResources.ipcSemSetID),
                     (void *)&ipcConHandle->ipcResources.ipcSemSetID);
     sqlt_system_error (SQLT_SQLCC, SQLT_SQLCC_IPC_AGENT_EXIT_LIST, 4,
                     sizeof(ipcConHandle->ipcResources.sendSemID),
                     (void *)&ipcConHandle->ipcResources.sendSemID);
     goto exit;
  }

exit:

  sqlt_cei_retcode(SQLT_SQLC, SQLT_SQLCC_IPC_AGENT_EXIT_LIST, 0);

}
                                                                  /*@ed13843mah*/
/* moved sqlccGetIPCs and sqlccFreeIPCs from sqlccipc_srv.C */    /*@f12409SAM*/

/*****************************************************************************
*
*  Function Name     = sqlccGetIPCs
*
*  Component Name    = SQLCC
*
*  Descriptive Name  = Allocate IPC resources for a specific connection
*
*  Func. Prototype   = sqlccGetIPCs(SQLCC_INITSTRUCT_T * initStruct)
*
*  Function          = Allocate the semaphore synchronization pair, and the
*                      shared memory communication buffer for the connection
*
*  Input             = 1) initStruct    : Pointer to the SQLCC_INITSTRUCT_T
*                                         control block.  The ipcResources
*                                         pointer must point to a valid
*                                         SQLCC_IPC_RESOURCES_T structure
*                                         which has the ipcSemSetID field
*                                         set.
*                      2) uid           : Effective userid of application
*
*  Output            = 1) initStruct    : Remainder of the ipcResources fields
*                                         are set
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_NO_ASL_HEAP                        @d11917tjv
*                      SQLCC_ERR_MEM_UNAVAIL                        @d16619tjv
*                      SQLCC_ERR_SYSTEM_ERROR
*
*****************************************************************************/

short sqlccGetIPCs( SQLCC_INITSTRUCT_T  * initStruct ,
                    SQLO_UID              uid )

{
   short        rc      = SQLCC_OK;
   int          sqlo_rc = SQLO_OK;

   struct   /* Flags for error recovery */
   {
      unsigned   mem_alloced       : 1;
      unsigned   sem_set_alloced   : 1;
      unsigned   send_sem_alloced  : 1;
      unsigned   recv_sem_alloced  : 1;
      unsigned   sync_sem_alloced  : 1;
      unsigned   clean_up          : 1;
   } errors = {0};

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCGETIPCS);

   /* Ensure that the shared memory handle and the receive and send */
   /* IDs in the ipcResources control block are cleared.            */

   sqlomset(&initStruct->ipcResources.ipcShmHandle, '\0',
             sizeof(initStruct->ipcResources.ipcShmHandle));
   sqlomset(&initStruct->ipcResources.recvSemID, '\0',
             sizeof(initStruct->ipcResources.recvSemID));
   sqlomset(&initStruct->ipcResources.sendSemID, '\0',
             sizeof(initStruct->ipcResources.sendSemID));
   initStruct->ipcResources.sendRecvSemOpen = FALSE;              /*@d45810tjv*/
   initStruct->ipcResources.noFreeShrMem = FALSE;                 //@d131103xun

   /* Allocate the shared memory communication buffer */

   sqlo_rc = sqlogmshr( SQLO_SHR_LOCAL_APL,
                        initStruct->ipcResources.ipcShmSize,
                        1,
                        &initStruct->ipcResources.ipcShmHandle);

   if (sqlo_rc != SQLO_OK)
   {
      if (sqlo_rc == SQLO_NOMEM)
      {
         sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 1,
                          sizeof(initStruct->ipcResources.ipcShmSize),
                          (void *)&initStruct->ipcResources.ipcShmSize);

         rc = SQLCC_ERR_NO_ASL_HEAP;
      }
      else
      {
         sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 1,
                          sizeof(initStruct->ipcResources.ipcShmSize),
                          (void *)&initStruct->ipcResources.ipcShmSize);

         sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 1,               /*@d18956glh*/
                         sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/  /*@d18956glh*/

         sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SHM_SIZE,              /*@d18956glh*/
                          (void *) &initStruct->ipcResources.ipcShmSize, /*@d18956glh*/
                          sizeof(initStruct->ipcResources.ipcShmSize),   /*@d18956glh*/
                          SQLT_FFSL_SEV);                                /*@d18956glh*/

         rc = SQLCC_ERR_SYSTEM_ERROR;
      }
      errors.clean_up = TRUE;
      goto exit;
   }
   errors.mem_alloced = TRUE;

   /* Give the application access to the shared memory set */

   sqlo_rc = sqloashm(&initStruct->ipcResources.ipcShmHandle, uid);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 3,             /*@d18956glh*/
                      sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/   /*@d18956glh*/

      sqlt_logerr_data((char *) SQLCC_FFDC_UID, (void *) &uid,      /*@d18956glh*/
                       sizeof(uid),SQLT_FFSL_SEV);                  /*@d18956glh*/

      rc = SQLCC_ERR_SYSTEM_ERROR;
      errors.clean_up = TRUE;
      goto exit;
   }

   /* Allocate the semaphore set for the conversation */

   sqlo_rc = sqloSSemSetCreate(SQLCC_IPC_NUM_SEMS_CONVERSATION,
                               &initStruct->ipcResources.ipcSemSetID);

   if (sqlo_rc != SQLO_OK)
   {
      if (sqlo_rc == SQLO_SSEM_INS_SYS)
      {
         rc = SQLCC_ERR_MEM_UNAVAIL;
      }
      else
      {
         sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 2,
                          sizeof(initStruct->ipcResources.ipcSemSetID),
                          (void *)&initStruct->ipcResources.ipcSemSetID);

         sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 5,                /*@d18956glh*/
                         sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/   /*@d18956glh*/

         sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,              /*@d18956glh*/
                          (void *) &initStruct->ipcResources.ipcSemSetID, /*@d18956glh*/
                          sizeof(initStruct->ipcResources.ipcSemSetID),   /*@d18956glh*/
                          SQLT_FFSL_SEV);                                 /*@d18956glh*/

         rc = SQLCC_ERR_SYSTEM_ERROR;
      }
      errors.clean_up = TRUE;
      goto exit;
   }
   errors.sem_set_alloced = TRUE;

   /* Grant access to the user to use this semaphore set */

   sqlo_rc = sqloSSemAccess(&initStruct->ipcResources.ipcSemSetID, uid);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 7,             /*@d18956glh*/
                      sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/   /*@d18956glh*/

      sqlt_logerr_data((char *) SQLCC_FFDC_UID, (void *) &uid,      /*@d18956glh*/
                       sizeof(uid), SQLT_FFSL_SEV);                 /*@d18956glh*/

      rc = SQLCC_ERR_SYSTEM_ERROR;
      errors.clean_up = TRUE;
      goto exit;
   }


   /* Allocate the receive synchronization semaphore */

   sqlo_rc = sqloSSemSetAllocate(&initStruct->ipcResources.ipcSemSetID,
                                 &initStruct->ipcResources.recvSemID);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 3, 8,
                       (void *)&initStruct->ipcResources.ipcSemSetID);

      switch (sqlo_rc)
      {
         case SQLO_SSEM_INS_SYS:
         case SQLO_SSEM_INS_MEM:
         case SQLO_SSEM_EXCEED_MAX:
         case SQLO_SSEM_INV_NUM:
            rc = SQLCC_ERR_MEM_UNAVAIL;
            break;
         default:
            sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 9,             /*@d18956glh*/
                            sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/   /*@d18956glh*/

            sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,           /*@d18956glh*/
                          (void *) &initStruct->ipcResources.ipcSemSetID, /*@d18956glh*/
                            sizeof(initStruct->ipcResources.ipcSemSetID), /*@d18956glh*/
                             SQLT_FFSL_SEV);                              /*@d18956glh*/

            rc = SQLCC_ERR_SYSTEM_ERROR;
            break;
      }
      errors.clean_up = TRUE;
      goto exit;
   }
   errors.recv_sem_alloced = TRUE;

   /* Allocate the send synchronization semaphore */

   sqlo_rc = sqloSSemSetAllocate(&initStruct->ipcResources.ipcSemSetID,
                                 &initStruct->ipcResources.sendSemID);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 4, 10,
                       (void *)&initStruct->ipcResources.ipcSemSetID);

      switch (sqlo_rc)
      {
         case SQLO_SSEM_INS_SYS:
         case SQLO_SSEM_INS_MEM:
         case SQLO_SSEM_EXCEED_MAX:
         case SQLO_SSEM_INV_NUM:
            rc = SQLCC_ERR_MEM_UNAVAIL;
            break;
         default:
             sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 11,           /*@d18956glh*/
                             sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/   /*@d18956glh*/

             sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,          /*@d18956glh*/
                          (void *) &initStruct->ipcResources.ipcSemSetID, /*@d18956glh*/
                           sizeof(initStruct->ipcResources.ipcSemSetID),  /*@d18956glh*/
                              SQLT_FFSL_SEV);                             /*@d18956glh*/

            rc = SQLCC_ERR_SYSTEM_ERROR;
            break;
      }
      errors.clean_up = TRUE;
      goto exit;
   }
   errors.send_sem_alloced = TRUE;

   /* Allocate the sync synchronization semaphore */

   sqlo_rc = sqloSSemSetAllocate(&initStruct->ipcResources.ipcSemSetID,
                                 &initStruct->ipcResources.syncSemID);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 5, 8,
                       (void *)&initStruct->ipcResources.ipcSemSetID);

      switch (sqlo_rc)
      {
         case SQLO_SSEM_INS_SYS:
         case SQLO_SSEM_INS_MEM:
         case SQLO_SSEM_EXCEED_MAX:
         case SQLO_SSEM_INV_NUM:
            rc = SQLCC_ERR_MEM_UNAVAIL;
            break;
         default:
           sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 13,             /*@d18956glh*/
                           sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/  /*@d18956glh*/

           sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,            /*@d18956glh*/
                          (void *) &initStruct->ipcResources.ipcSemSetID, /*@d18956glh*/
                            sizeof(initStruct->ipcResources.ipcSemSetID), /*@d18956glh*/
                            SQLT_FFSL_SEV);                               /*@d18956glh*/

            rc = SQLCC_ERR_SYSTEM_ERROR;
            break;
      }
      errors.clean_up = TRUE;
      goto exit;
   }
   errors.sync_sem_alloced = TRUE;

exit:

   /* Perform cleanup recovery if required */

   if (errors.clean_up)
   {
      /* If the shared memory set was allocated free it */

      if (errors.mem_alloced)
      {
         sqlo_rc = sqlofshr(&initStruct->ipcResources.ipcShmHandle,
                            SQLO_SHR_NOCHECK);
         if (sqlo_rc != SQLO_OK)
         {
            sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 6,
                             sizeof(initStruct->ipcResources.ipcShmHandle),
                             (void *)&initStruct->ipcResources.ipcShmHandle);

        sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 15,                /*@d18956glh*/
                        sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/   /*@d18956glh*/

        sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SHM_HANDLE,              /*@d18956glh*/
                         (void *) &initStruct->ipcResources.ipcShmHandle, /*@d18956glh*/
                         sizeof(initStruct->ipcResources.ipcShmHandle),   /*@d18956glh*/
                         SQLT_FFSL_SEV);                                  /*@d18956glh*/

            rc = SQLCC_ERR_SYSTEM_ERROR;
         }
         sqlomset(&initStruct->ipcResources.ipcShmHandle, '\0',
                  sizeof(initStruct->ipcResources.ipcShmHandle));
      }

      /* If the receive semaphore was allocated free it */

      if (errors.recv_sem_alloced)
      {
         sqlo_rc = sqloSSemSetFree(&initStruct->ipcResources.ipcSemSetID,
                                   &initStruct->ipcResources.recvSemID);
         if (sqlo_rc != SQLO_OK)
         {
            sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 7, 8,
                           (void *)&initStruct->ipcResources.ipcSemSetID, 8,
                           (void *)&initStruct->ipcResources.recvSemID);

         sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 17,               /*@d18956glh*/
                         sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/  /*@d18956glh*/

         sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,              /*@d18956glh*/
                          (void *) &initStruct->ipcResources.ipcSemSetID, /*@d18956glh*/
                          8, SQLT_FFSL_SEV);                              /*@d18956glh*/

         sqlt_logerr_data((char *) SQLCC_FFDC_RECV_SEMID,                 /*@d18956glh*/
                          (void *) &initStruct->ipcResources.recvSemID,   /*@d18956glh*/
                          8, SQLT_FFSL_SEV);                              /*@d18956glh*/

            rc = SQLCC_ERR_SYSTEM_ERROR;
         }
         sqlomset(&initStruct->ipcResources.recvSemID, '\0',
                  sizeof(initStruct->ipcResources.recvSemID));
      }

      /* If the send semaphore was allocated free it */

      if (errors.send_sem_alloced)
      {
         sqlo_rc = sqloSSemSetFree(&initStruct->ipcResources.ipcSemSetID,
                                   &initStruct->ipcResources.sendSemID);
         if (sqlo_rc != SQLO_OK)
         {
            sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 8, 8,
                           (void *)&initStruct->ipcResources.ipcSemSetID, 8,
                           (void *)&initStruct->ipcResources.sendSemID);

         sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 20,               /*@d18956glh*/
                         sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/   /*@d18956glh*/

         sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,              /*@d18956glh*/
                          (void *) &initStruct->ipcResources.ipcSemSetID, /*@d18956glh*/
                          8, SQLT_FFSL_SEV);                              /*@d18956glh*/

         sqlt_logerr_data((char *) SQLCC_FFDC_SEND_SEMID,                 /*@d18956glh*/
                          (void *) &initStruct->ipcResources.sendSemID,   /*@d18956glh*/
                          8, SQLT_FFSL_SEV);                              /*@d18956glh*/

            rc = SQLCC_ERR_SYSTEM_ERROR;
         }
         sqlomset(&initStruct->ipcResources.sendSemID, '\0',
                  sizeof(initStruct->ipcResources.sendSemID));
      }

      /* If the synchronization semaphore was allocated, free it */

      if (errors.sync_sem_alloced)
      {
         sqlo_rc = sqloSSemSetFree(&initStruct->ipcResources.ipcSemSetID,
                                   &initStruct->ipcResources.syncSemID);
         if (sqlo_rc != SQLO_OK)
         {
            sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 8, 8,
                           (void *)&initStruct->ipcResources.ipcSemSetID, 8,
                           (void *)&initStruct->ipcResources.syncSemID);

         sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 23,               /*@d18956glh*/
                         sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/   /*@d18956glh*/

         sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,              /*@d18956glh*/
                          (void *) &initStruct->ipcResources.ipcSemSetID, /*@d18956glh*/
                          8, SQLT_FFSL_SEV);                              /*@d18956glh*/

         sqlt_logerr_data((char *) SQLCC_FFDC_SYNC_SEMID,                 /*@d18956glh*/
                          (void *) &initStruct->ipcResources.syncSemID,   /*@d18956glh*/
                          8, SQLT_FFSL_SEV);                              /*@d18956glh*/

            rc = SQLCC_ERR_SYSTEM_ERROR;
         }
         sqlomset(&initStruct->ipcResources.syncSemID, '\0',
                  sizeof(initStruct->ipcResources.syncSemID));
      }

      /* If the semaphore set was allocated free it */

      if (errors.sem_set_alloced)
      {
         sqlo_rc = sqloSSemSetRemove(&initStruct->ipcResources.ipcSemSetID);

         if (sqlo_rc != SQLO_OK)
         {
            sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 8, 8,
                          (void *)&initStruct->ipcResources.ipcSemSetID);

         sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCGETIPCS, 26,               /*@d18956glh*/
                         sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN); /*@d106920jh*/  /*@d18956glh*/

         sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,              /*@d18956glh*/
                          (void *) &initStruct->ipcResources.ipcSemSetID, /*@d18956glh*/
                          8, SQLT_FFSL_SEV);                              /*@d18956glh*/

            rc = SQLCC_ERR_SYSTEM_ERROR;
         }
         sqlomset(&initStruct->ipcResources.ipcSemSetID, '\0',
                  sizeof(initStruct->ipcResources.ipcSemSetID));
      }

   }

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCGETIPCS, rc);
   return(rc);
}
#endif  //BLD_CLIENT_ONLY

/*****************************************************************************
*
*  Function Name     = sqlccFreeIPCs
*
*  Component Name    = SQLCC
*
*  Descriptive Name  = Allocate IPC resources for a specific connection
*
*  Func. Prototype   = sqlccGetIPCs(SQLCC_INITSTRUCT_T * initStruct)
*
*  Function          = Free the semaphore synchronization pair, and the
*                      shared memory communication buffer for the connection
*
*  Input             = 1) initStruct    : Pointer to the SQLCC_INITSTRUCT_T
*                                         control block.  The ipcResources
*                                         pointer must point to a valid
*                                         SQLCC_IPC_RESOURCES_T structure
*                                         which has the ipcSemSetID field
*                                         set.
*                      2) isStoredProc  : Whether this is cleanup for
*                                         a stored proc. Since > 1 process
*                                         may clean up an sps ipcs, we
*                                         refrain from error logging
*                      3) isSysc        : is this a stored proc cleanup
*                                         driven by the db2sysc?
*
*  Output            = 2) initStruct    : Remainder of the ipcResources fields
*                                         are cleared
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_SYSTEM_ERROR
*
*****************************************************************************/

short sqlccFreeIPCs( SQLCC_INITSTRUCT_T  * initStruct,
                     char                  isStoredProc,           //@d219059swm
                     char                  isSysc )                //@d219059swm

{
   short        rc      = SQLCC_OK;
   int          sqlo_rc = SQLO_OK;

   sqlt_fnc_entry(SQLT_SQLCC, SQLT_SQLCCFREEIPCS);

   /* Free the shared memory set and clear the ipcShmHandle */
   if(isSysc == FALSE)                                             //@d219059swm
   {
     sqlo_rc = sqlofshr(&initStruct->ipcResources.ipcShmHandle,
                        SQLO_SHR_NOCHECK);

     if ((sqlo_rc != SQLO_OK) &&
         (sqlo_rc != SQLO_INV_MEM))
     {
        sqlt_trace_error(SQLT_SQLCC, SQLT_SQLCCFREEIPCS, 1,
                         sizeof(initStruct->ipcResources.ipcShmHandle),
                         (void *)&initStruct->ipcResources.ipcShmHandle);

        if(!isStoredProc)
        {
          sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCFREEIPCS, 1,
                          sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN);

          sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SHM_HANDLE,
                           (void *) &initStruct->ipcResources.ipcShmHandle,
                           sizeof(initStruct->ipcResources.ipcShmHandle),
                           SQLT_FFSL_SEV);
       }
       rc = SQLCC_ERR_SYSTEM_ERROR;
     }
   }

   initStruct->ipcResources.ipcShmHandle.pAddress = NULL;           //@d80385xun

   /* Free the receive semaphore and clear the recvSemID */

   sqlo_rc = sqloSSemSetFree(&initStruct->ipcResources.ipcSemSetID,
                             &initStruct->ipcResources.recvSemID);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCFREEIPCS, 2, 8,
                     (void *)&initStruct->ipcResources.ipcSemSetID, 8,
                     (void *)&initStruct->ipcResources.recvSemID);

      if(!isStoredProc)                                           //@db219059swm
      {
        sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCFREEIPCS, 3,
                        sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN);

        sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,
                         (void *) &initStruct->ipcResources.ipcSemSetID,
                         8, SQLT_FFSL_SEV);

        sqlt_logerr_data((char *) SQLCC_FFDC_RECV_SEMID,
                         (void *) &initStruct->ipcResources.recvSemID,
                         8, SQLT_FFSL_SEV);
      }                                                           //@de219059swm

      rc = SQLCC_ERR_SYSTEM_ERROR;
   }

   sqlomset(&initStruct->ipcResources.recvSemID, '\0',
            sizeof(initStruct->ipcResources.recvSemID));

   /* Free the send semaphore and clear the sendSemID */

   sqlo_rc = sqloSSemSetFree(&initStruct->ipcResources.ipcSemSetID,
                             &initStruct->ipcResources.sendSemID);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCFREEIPCS, 3, 8,
                     (void *)&initStruct->ipcResources.ipcSemSetID, 8,
                     (void *)&initStruct->ipcResources.sendSemID);

      if(!isStoredProc)                                           //@db219059swm
      {
        sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCFREEIPCS, 6,
                        sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN);

        sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,
                         (void *) &initStruct->ipcResources.ipcSemSetID,
                         8, SQLT_FFSL_SEV);

        sqlt_logerr_data((char *) SQLCC_FFDC_SEND_SEMID,
                         (void *) &initStruct->ipcResources.sendSemID,
                         8, SQLT_FFSL_SEV);
      }                                                           //@db219059swm

      rc = SQLCC_ERR_SYSTEM_ERROR;
   }

   sqlomset(&initStruct->ipcResources.sendSemID, '\0',
            sizeof(initStruct->ipcResources.sendSemID));

   /* Free the sync semaphore */

   sqlo_rc = sqloSSemSetFree(&initStruct->ipcResources.ipcSemSetID,
                             &initStruct->ipcResources.syncSemID);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_fnc_data2(SQLT_SQLCC, SQLT_SQLCCFREEIPCS, 4, 8,
                     (void *)&initStruct->ipcResources.ipcSemSetID, 8,
                     (void *)&initStruct->ipcResources.syncSemID);

      if(!isStoredProc)                                           //@db219059swm
      {
        sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCFREEIPCS, 9,
                        sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN);

        sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,
                         (void *) &initStruct->ipcResources.ipcSemSetID,
                         8, SQLT_FFSL_SEV);

        sqlt_logerr_data((char *) SQLCC_FFDC_SYNC_SEMID,
                         (void *) &initStruct->ipcResources.syncSemID,
                         8, SQLT_FFSL_SEV);
      }                                                           //@de219059swm

      rc = SQLCC_ERR_SYSTEM_ERROR;
   }

   sqlomset(&initStruct->ipcResources.syncSemID, '\0',
            sizeof(initStruct->ipcResources.syncSemID));

   /* Remove the semaphore set */

   sqlo_rc = sqloSSemSetRemove(&initStruct->ipcResources.ipcSemSetID);

   if (sqlo_rc != SQLO_OK)
   {
      sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCCFREEIPCS, 5, 8,
                    (void *)&initStruct->ipcResources.ipcSemSetID);

      if(!isStoredProc)                                           //@db219059swm
      {
        sqlt_logerr_zrc(SQLT_SQLCC, SQLT_SQLCCFREEIPCS, 12,
                        sqlo_rc, SQLT_FFSL_SEV | SQLT_ADMIN);

        sqlt_logerr_data((char *) SQLCC_FFDC_IPC_SEM_SETID,
                         (void *) &initStruct->ipcResources.ipcSemSetID,
                         8, SQLT_FFSL_SEV);
      }                                                           //@db219059swm

      rc = SQLCC_ERR_SYSTEM_ERROR;
   }
   sqlomset(&initStruct->ipcResources.ipcSemSetID, '\0',
            sizeof(initStruct->ipcResources.ipcSemSetID));

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCCFREEIPCS, rc);
   return(rc);
}

                                                                       //@db111814bpg
/*****************************************************************************
*
*  Function Name     = sqlccIPCSignalDariEnd
*
*  Component Name    = SQLCC
*
*  Descriptive Name  = manipulate the IPC semaphores to signal a dari to die
*
*  Func. Prototype   = sqlccIPCSignalDariEnd( SQLCC_IPC_RESOURCES_T *  ipcResources )
*
*  Function          = attach to shared memory and set the receive term flag,
*                      then post the send and receive semaphores to wake up the dair
*
*  Input             = 1) ipcResources  : Pointer to the SQLCC_IPC_RESOURCES
*                                         structure.
*
*  Output            = none
*
*  Normal Return     = SQLCC_OK
*
*  Error Return      = SQLCC_ERR_SYSTEM_ERROR
*
*****************************************************************************/

short  sqlccIPCSignalDariEnd( SQLCC_IPC_RESOURCES_T *  ipcResources )

{
   SQLCC_IPC_ASL_HDR_T *  aslCommHeader;
   short  rc = SQLCC_OK;
   int  sqlo_rc;
   bool   connect_to_shmem = false;                                //@d143023swm
   bool   recv_semaphore_open  = false;                            //@d143023swm
   bool   send_semaphore_open  = false;                            //@d143023swm

   sqlt_fnc_entry(SQLT_SQLCC2, SQLT_SQLCC_IPC_SIGNAL_DARI_END );

   // Attach to the shared memory communication buffer

   sqlo_rc = sqlocshr(&ipcResources->ipcShmHandle);
   if (sqlo_rc != SQLO_OK)
   {
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
      sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 1,
                    sizeof(&ipcResources->ipcShmHandle),
                    (void *)&ipcResources->ipcShmHandle);
      goto exit;
   }
   connect_to_shmem = true;                                        //@d143023swm

   // get the comm header pointer - set the recv term flag
   // so the dari knows to die

   aslCommHeader = (SQLCC_IPC_ASL_HDR_T *)
               sqloGetSetRawAddress(&ipcResources->ipcShmHandle);
   aslCommHeader -> recvTermFlag = TRUE;

   /* Open the receive semaphore */
   sqlo_rc = sqloSSemOpen(&ipcResources->ipcSemSetID,
                          &ipcResources->recvSemID);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error( SQLT_SQLCC2, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 2,
         sizeof(sqlo_rc), &sqlo_rc);
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
      goto exit;
   }
   recv_semaphore_open = true;                                     //@d143023swm

   /* post the receive semaphore to wake up the DARI */
   sqlo_rc = sqloSSemV(&ipcResources->ipcSemSetID,
                       &ipcResources->recvSemID,
                       0);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error( SQLT_SQLCC2, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 3,
         sizeof(sqlo_rc), &sqlo_rc);
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
      goto exit;
   }

   /* Open the send semaphore */
   sqlo_rc = sqloSSemOpen(&ipcResources->ipcSemSetID,
                          &ipcResources->sendSemID);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error( SQLT_SQLCC2, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 4,
         sizeof(sqlo_rc), &sqlo_rc);
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
      goto exit;
   }
   send_semaphore_open = true;                                     //@d143023swm

   /* post the send semaphore which will cause the DARI to fail */
   sqlo_rc = sqloSSemV(&ipcResources->ipcSemSetID,
                       &ipcResources->sendSemID,
                       0);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error( SQLT_SQLCC2, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 5,
         sizeof(sqlo_rc), &sqlo_rc);
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
   }


exit:
   if(send_semaphore_open == true)                                //@db143023swm
   {
     sqlo_rc = sqloSSemClose(&ipcResources->ipcSemSetID,
                             &ipcResources->sendSemID);

     if (sqlo_rc != SQLO_OK)
     {
        if(rc == SQL_RC_OK)
        {
          rc = SQLCC_ERR_COMM_REQUEST_FAILED;
        }
        sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 40,
                      sizeof(sqlo_rc), &sqlo_rc);
     }
   }

   if(recv_semaphore_open == true)
   {
     sqlo_rc = sqloSSemClose(&ipcResources->ipcSemSetID,
                             &ipcResources->recvSemID);

     if (sqlo_rc != SQLO_OK)
     {
        if(rc == SQL_RC_OK)
        {
          rc = SQLCC_ERR_COMM_REQUEST_FAILED;
        }
        sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 45,
                      sizeof(sqlo_rc), &sqlo_rc);
     }
   }

   if(connect_to_shmem == true)
   {
     sqlo_rc = sqlodshr(&ipcResources->ipcShmHandle);
     if (sqlo_rc != SQLO_OK)
     {
        if(rc == SQL_RC_OK)
        {
          rc = SQLCC_ERR_COMM_REQUEST_FAILED;
        }
        sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 30,
                      sizeof(sqlo_rc), &sqlo_rc);
     }
   }                                                              //@de143023swm

   sqlt_fnc_retcode(SQLT_SQLCC, SQLT_SQLCC_IPC_SIGNAL_DARI_END, rc);
   return rc;
}                                                                      //@de111814bpg


   bool   recv_semaphore_open  = false;                            //@d143023swm
   bool   send_semaphore_open  = false;                            //@d143023swm
   sqlt_fnc_entry(SQLT_SQLCC2, SQLT_SQLCC_IPC_SIGNAL_DARI_END );
   // Attach to the shared memory communication buffer
   sqlo_rc = sqlocshr(&ipcResources->ipcShmHandle);
   if (sqlo_rc != SQLO_OK)
   {
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
      sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 1,
                    sizeof(&ipcResources->ipcShmHandle),
                    (void *)&ipcResources->ipcShmHandle);
      goto exit;
   }
   connect_to_shmem = true;                                        //@d143023swm
   // get the comm header pointer - set the recv term flag
   // so the dari knows to die
   aslCommHeader = (SQLCC_IPC_ASL_HDR_T *)
               sqloGetSetRawAddress(&ipcResources->ipcShmHandle);
   aslCommHeader -> recvTermFlag = TRUE;
   /* Open the receive semaphore */
   sqlo_rc = sqloSSemOpen(&ipcResources->ipcSemSetID,
                          &ipcResources->recvSemID);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error( SQLT_SQLCC2, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 2,
         sizeof(sqlo_rc), &sqlo_rc);
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
      goto exit;
   }
   recv_semaphore_open = true;                                     //@d143023swm
   /* post the receive semaphore to wake up the DARI */
   sqlo_rc = sqloSSemV(&ipcResources->ipcSemSetID,
                       &ipcResources->recvSemID,
                       0);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error( SQLT_SQLCC2, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 3,
         sizeof(sqlo_rc), &sqlo_rc);
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
      goto exit;
   }
   /* Open the send semaphore */
   sqlo_rc = sqloSSemOpen(&ipcResources->ipcSemSetID,
                          &ipcResources->sendSemID);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error( SQLT_SQLCC2, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 4,
         sizeof(sqlo_rc), &sqlo_rc);
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
      goto exit;
   }
   send_semaphore_open = true;                                     //@d143023swm
   /* post the send semaphore which will cause the DARI to fail */
   sqlo_rc = sqloSSemV(&ipcResources->ipcSemSetID,
                       &ipcResources->sendSemID,
                       0);
   if (sqlo_rc != SQLO_OK)
   {
      sqlt_trace_error( SQLT_SQLCC2, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 5,
         sizeof(sqlo_rc), &sqlo_rc);
      rc = SQLCC_ERR_COMM_REQUEST_FAILED;
   }

exit:
   if(send_semaphore_open == true)                                //@db143023swm
   {
     sqlo_rc = sqloSSemClose(&ipcResources->ipcSemSetID,
                             &ipcResources->sendSemID);
     if (sqlo_rc != SQLO_OK)
     {
        if(rc == SQL_RC_OK)
        {
          rc = SQLCC_ERR_COMM_REQUEST_FAILED;
        }
        sqlt_fnc_data(SQLT_SQLCC, SQLT_SQLCC_IPC_SIGNAL_DARI_END, 40,
                      sizeof(sqlo_rc), &sqlo_rc);
     }
   }
   if(recv_semaphore_open == true)
   {
     sqlo_rc = sqloSSemClose(&ipcResources->ipcSemSetID,
